from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class SH(DefinedNamespace):
    """
    W3C Shapes Constraint Language (SHACL) Vocabulary

    This vocabulary defines terms used in SHACL, the W3C Shapes Constraint Language.

    Generated from: https://www.w3.org/ns/shacl.ttl
    Date: 2020-05-26 14:20:08.041103

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    alternativePath: URIRef  # The (single) value of this property must be a list of path elements, representing the elements of alternative paths.
    annotationProperty: URIRef  # The annotation property that shall be set.
    annotationValue: URIRef  # The (default) values of the annotation property.
    annotationVarName: URIRef  # The name of the SPARQL variable from the SELECT clause that shall be used for the values.
    ask: URIRef  # The SPARQL ASK query to execute.
    closed: URIRef  # If set to true then the shape is closed.
    condition: URIRef  # The shapes that the focus nodes need to conform to before a rule is executed on them.
    conforms: URIRef  # True if the validation did not produce any validation results, and false otherwise.
    construct: URIRef  # The SPARQL CONSTRUCT query to execute.
    datatype: URIRef  # Specifies an RDF datatype that all value nodes must have.
    deactivated: URIRef  # If set to true then all nodes conform to this.
    declare: URIRef  # Links a resource with its namespace prefix declarations.
    defaultValue: URIRef  # A default value for a property, for example for user interface tools to pre-populate input fields.
    description: URIRef  # Human-readable descriptions for the property in the context of the surrounding shape.
    detail: URIRef  # Links a result with other results that provide more details, for example to describe violations against nested shapes.
    disjoint: URIRef  # Specifies a property where the set of values must be disjoint with the value nodes.
    entailment: URIRef  # An entailment regime that indicates what kind of inferencing is required by a shapes graph.
    equals: URIRef  # Specifies a property that must have the same values as the value nodes.
    expression: URIRef  # The node expression that must return true for the value nodes.
    filterShape: (
        URIRef  # The shape that all input nodes of the expression need to conform to.
    )
    flags: (
        URIRef  # An optional flag to be used with regular expression pattern matching.
    )
    focusNode: URIRef  # The focus node that was validated when the result was produced.
    group: URIRef  # Can be used to link to a property group to indicate that a property shape belongs to a group of related property shapes.
    hasValue: URIRef  # Specifies a value that must be among the value nodes.
    ignoredProperties: URIRef  # An optional RDF list of properties that are also permitted in addition to those explicitly enumerated via sh:property/sh:path.
    intersection: URIRef  # A list of node expressions that shall be intersected.
    inversePath: URIRef  # The (single) value of this property represents an inverse path (object to subject).
    js: URIRef  # Constraints expressed in JavaScript.
    jsFunctionName: URIRef  # The name of the JavaScript function to execute.
    jsLibrary: URIRef  # Declares which JavaScript libraries are needed to execute this.
    jsLibraryURL: URIRef  # Declares the URLs of a JavaScript library. This should be the absolute URL of a JavaScript file. Implementations may redirect those to local files.
    labelTemplate: URIRef  # Outlines how human-readable labels of instances of the associated Parameterizable shall be produced. The values can contain {?paramName} as placeholders for the actual values of the given parameter.
    languageIn: (
        URIRef  # Specifies a list of language tags that all value nodes must have.
    )
    lessThan: URIRef  # Specifies a property that must have smaller values than the value nodes.
    lessThanOrEquals: URIRef  # Specifies a property that must have smaller or equal values than the value nodes.
    maxCount: (
        URIRef  # Specifies the maximum number of values in the set of value nodes.
    )
    maxExclusive: URIRef  # Specifies the maximum exclusive value of each value node.
    maxInclusive: URIRef  # Specifies the maximum inclusive value of each value node.
    maxLength: URIRef  # Specifies the maximum string length of each value node.
    message: URIRef  # A human-readable message (possibly with placeholders for variables) explaining the cause of the result.
    minCount: (
        URIRef  # Specifies the minimum number of values in the set of value nodes.
    )
    minExclusive: URIRef  # Specifies the minimum exclusive value of each value node.
    minInclusive: URIRef  # Specifies the minimum inclusive value of each value node.
    minLength: URIRef  # Specifies the minimum string length of each value node.
    name: URIRef  # Human-readable labels for the property in the context of the surrounding shape.
    namespace: URIRef  # The namespace associated with a prefix in a prefix declaration.
    node: URIRef  # Specifies the node shape that all value nodes must conform to.
    nodeKind: URIRef  # Specifies the node kind (e.g. IRI or literal) each value node.
    nodeValidator: URIRef  # The validator(s) used to evaluate a constraint in the context of a node shape.
    nodes: URIRef  # The node expression producing the input nodes of a filter shape expression.
    object: (
        URIRef  # An expression producing the nodes that shall be inferred as objects.
    )
    oneOrMorePath: URIRef  # The (single) value of this property represents a path that is matched one or more times.
    optional: URIRef  # Indicates whether a parameter is optional.
    order: URIRef  # Specifies the relative order of this compared to its siblings. For example use 0 for the first, 1 for the second.
    parameter: URIRef  # The parameters of a function or constraint component.
    path: URIRef  # Specifies the property path of a property shape.
    pattern: URIRef  # Specifies a regular expression pattern that the string representations of the value nodes must match.
    predicate: URIRef  # An expression producing the properties that shall be inferred as predicates.
    prefix: URIRef  # The prefix of a prefix declaration.
    prefixes: URIRef  # The prefixes that shall be applied before parsing the associated SPARQL query.
    property: URIRef  # Links a shape to its property shapes.
    propertyValidator: URIRef  # The validator(s) used to evaluate a constraint in the context of a property shape.
    qualifiedMaxCount: (
        URIRef  # The maximum number of value nodes that can conform to the shape.
    )
    qualifiedMinCount: (
        URIRef  # The minimum number of value nodes that must conform to the shape.
    )
    qualifiedValueShape: (
        URIRef  # The shape that a specified number of values must conform to.
    )
    qualifiedValueShapesDisjoint: URIRef  # Can be used to mark the qualified value shape to be disjoint with its sibling shapes.
    result: URIRef  # The validation results contained in a validation report.
    resultAnnotation: URIRef  # Links a SPARQL validator with zero or more sh:ResultAnnotation instances, defining how to derive additional result properties based on the variables of the SELECT query.
    resultMessage: URIRef  # Human-readable messages explaining the cause of the result.
    resultPath: URIRef  # The path of a validation result, based on the path of the validated property shape.
    resultSeverity: URIRef  # The severity of the result, e.g. warning.
    returnType: (
        URIRef  # The expected type of values returned by the associated function.
    )
    rule: URIRef  # The rules linked to a shape.
    select: URIRef  # The SPARQL SELECT query to execute.
    severity: URIRef  # Defines the severity that validation results produced by a shape must have. Defaults to sh:Violation.
    shapesGraph: (
        URIRef  # Shapes graphs that should be used when validating this data graph.
    )
    shapesGraphWellFormed: URIRef  # If true then the validation engine was certain that the shapes graph has passed all SHACL syntax requirements during the validation process.
    sourceConstraint: (
        URIRef  # The constraint that was validated when the result was produced.
    )
    sourceConstraintComponent: (
        URIRef  # The constraint component that is the source of the result.
    )
    sourceShape: URIRef  # The shape that is was validated when the result was produced.
    sparql: URIRef  # Links a shape with SPARQL constraints.
    subject: URIRef  # An expression producing the resources that shall be inferred as subjects.
    suggestedShapesGraph: URIRef  # Suggested shapes graphs for this ontology. The values of this property may be used in the absence of specific sh:shapesGraph statements.
    target: URIRef  # Links a shape to a target specified by an extension language, for example instances of sh:SPARQLTarget.
    targetClass: URIRef  # Links a shape to a class, indicating that all instances of the class must conform to the shape.
    targetNode: URIRef  # Links a shape to individual nodes, indicating that these nodes must conform to the shape.
    targetObjectsOf: URIRef  # Links a shape to a property, indicating that all all objects of triples that have the given property as their predicate must conform to the shape.
    targetSubjectsOf: URIRef  # Links a shape to a property, indicating that all subjects of triples that have the given property as their predicate must conform to the shape.
    union: URIRef  # A list of node expressions that shall be used together.
    uniqueLang: URIRef  # Specifies whether all node values must have a unique (or no) language tag.
    update: URIRef  # The SPARQL UPDATE to execute.
    validator: URIRef  # The validator(s) used to evaluate constraints of either node or property shapes.
    value: URIRef  # An RDF node that has caused the result.
    xone: URIRef  # Specifies a list of shapes so that the value nodes must conform to exactly one of the shapes.
    zeroOrMorePath: URIRef  # The (single) value of this property represents a path that is matched zero or more times.
    zeroOrOnePath: URIRef  # The (single) value of this property represents a path that is matched zero or one times.

    # http://www.w3.org/2000/01/rdf-schema#Class
    AbstractResult: URIRef  # The base class of validation results, typically not instantiated directly.
    ConstraintComponent: URIRef  # The class of constraint components.
    Function: URIRef  # The class of SHACL functions.
    JSConstraint: URIRef  # The class of constraints backed by a JavaScript function.
    JSExecutable: URIRef  # Abstract base class of resources that declare an executable JavaScript.
    JSFunction: URIRef  # The class of SHACL functions that execute a JavaScript function when called.
    JSLibrary: URIRef  # Represents a JavaScript library, typically identified by one or more URLs of files to include.
    JSRule: URIRef  # The class of SHACL rules expressed using JavaScript.
    JSTarget: URIRef  # The class of targets that are based on JavaScript functions.
    JSTargetType: URIRef  # The (meta) class for parameterizable targets that are based on JavaScript functions.
    JSValidator: URIRef  # A SHACL validator based on JavaScript. This can be used to declare SHACL constraint components that perform JavaScript-based validation when used.
    NodeKind: URIRef  # The class of all node kinds, including sh:BlankNode, sh:IRI, sh:Literal or the combinations of these: sh:BlankNodeOrIRI, sh:BlankNodeOrLiteral, sh:IRIOrLiteral.
    NodeShape: URIRef  # A node shape is a shape that specifies constraint that need to be met with respect to focus nodes.
    Parameter: URIRef  # The class of parameter declarations, consisting of a path predicate and (possibly) information about allowed value type, cardinality and other characteristics.
    Parameterizable: URIRef  # Superclass of components that can take parameters, especially functions and constraint components.
    PrefixDeclaration: URIRef  # The class of prefix declarations, consisting of pairs of a prefix with a namespace.
    PropertyGroup: URIRef  # Instances of this class represent groups of property shapes that belong together.
    PropertyShape: URIRef  # A property shape is a shape that specifies constraints on the values of a focus node for a given property or path.
    ResultAnnotation: URIRef  # A class of result annotations, which define the rules to derive the values of a given annotation property as extra values for a validation result.
    Rule: URIRef  # The class of SHACL rules. Never instantiated directly.
    SPARQLAskExecutable: (
        URIRef  # The class of SPARQL executables that are based on an ASK query.
    )
    SPARQLAskValidator: URIRef  # The class of validators based on SPARQL ASK queries. The queries are evaluated for each value node and are supposed to return true if the given node conforms.
    SPARQLConstraint: URIRef  # The class of constraints based on SPARQL SELECT queries.
    SPARQLConstructExecutable: (
        URIRef  # The class of SPARQL executables that are based on a CONSTRUCT query.
    )
    SPARQLExecutable: URIRef  # The class of resources that encapsulate a SPARQL query.
    SPARQLFunction: (
        URIRef  # A function backed by a SPARQL query - either ASK or SELECT.
    )
    SPARQLRule: URIRef  # The class of SHACL rules based on SPARQL CONSTRUCT queries.
    SPARQLSelectExecutable: (
        URIRef  # The class of SPARQL executables based on a SELECT query.
    )
    SPARQLSelectValidator: URIRef  # The class of validators based on SPARQL SELECT queries. The queries are evaluated for each focus node and are supposed to produce bindings for all focus nodes that do not conform.
    SPARQLTarget: URIRef  # The class of targets that are based on SPARQL queries.
    SPARQLTargetType: URIRef  # The (meta) class for parameterizable targets that are based on SPARQL queries.
    SPARQLUpdateExecutable: (
        URIRef  # The class of SPARQL executables based on a SPARQL UPDATE.
    )
    Severity: URIRef  # The class of validation result severity levels, including violation and warning levels.
    Shape: URIRef  # A shape is a collection of constraints that may be targeted for certain nodes.
    Target: URIRef  # The base class of targets such as those based on SPARQL queries.
    TargetType: URIRef  # The (meta) class for parameterizable targets.	Instances of this are instantiated as values of the sh:target property.
    TripleRule: URIRef  # A rule based on triple (subject, predicate, object) pattern.
    ValidationReport: URIRef  # The class of SHACL validation reports.
    ValidationResult: URIRef  # The class of validation results.
    Validator: URIRef  # The class of validators, which provide instructions on how to process a constraint definition. This class serves as base class for the SPARQL-based validators and other possible implementations.

    # http://www.w3.org/2000/01/rdf-schema#Resource
    this: URIRef  # A node expression that represents the current focus node.

    # http://www.w3.org/ns/shacl#ConstraintComponent
    AndConstraintComponent: URIRef  # A constraint component that can be used to test whether a value node conforms to all members of a provided list of shapes.
    ClassConstraintComponent: URIRef  # A constraint component that can be used to verify that each value node is an instance of a given type.
    ClosedConstraintComponent: URIRef  # A constraint component that can be used to indicate that focus nodes must only have values for those properties that have been explicitly enumerated via sh:property/sh:path.
    DatatypeConstraintComponent: URIRef  # A constraint component that can be used to restrict the datatype of all value nodes.
    DisjointConstraintComponent: URIRef  # A constraint component that can be used to verify that the set of value nodes is disjoint with the the set of nodes that have the focus node as subject and the value of a given property as predicate.
    EqualsConstraintComponent: URIRef  # A constraint component that can be used to verify that the set of value nodes is equal to the set of nodes that have the focus node as subject and the value of a given property as predicate.
    ExpressionConstraintComponent: URIRef  # A constraint component that can be used to verify that a given node expression produces true for all value nodes.
    HasValueConstraintComponent: URIRef  # A constraint component that can be used to verify that one of the value nodes is a given RDF node.
    InConstraintComponent: URIRef  # A constraint component that can be used to exclusively enumerate the permitted value nodes.
    JSConstraintComponent: URIRef  # A constraint component with the parameter sh:js linking to a sh:JSConstraint containing a sh:script.
    LanguageInConstraintComponent: URIRef  # A constraint component that can be used to enumerate language tags that all value nodes must have.
    LessThanConstraintComponent: URIRef  # A constraint component that can be used to verify that each value node is smaller than all the nodes that have the focus node as subject and the value of a given property as predicate.
    LessThanOrEqualsConstraintComponent: URIRef  # A constraint component that can be used to verify that every value node is smaller than all the nodes that have the focus node as subject and the value of a given property as predicate.
    MaxCountConstraintComponent: URIRef  # A constraint component that can be used to restrict the maximum number of value nodes.
    MaxExclusiveConstraintComponent: URIRef  # A constraint component that can be used to restrict the range of value nodes with a maximum exclusive value.
    MaxInclusiveConstraintComponent: URIRef  # A constraint component that can be used to restrict the range of value nodes with a maximum inclusive value.
    MaxLengthConstraintComponent: URIRef  # A constraint component that can be used to restrict the maximum string length of value nodes.
    MinCountConstraintComponent: URIRef  # A constraint component that can be used to restrict the minimum number of value nodes.
    MinExclusiveConstraintComponent: URIRef  # A constraint component that can be used to restrict the range of value nodes with a minimum exclusive value.
    MinInclusiveConstraintComponent: URIRef  # A constraint component that can be used to restrict the range of value nodes with a minimum inclusive value.
    MinLengthConstraintComponent: URIRef  # A constraint component that can be used to restrict the minimum string length of value nodes.
    NodeConstraintComponent: URIRef  # A constraint component that can be used to verify that all value nodes conform to the given node shape.
    NodeKindConstraintComponent: URIRef  # A constraint component that can be used to restrict the RDF node kind of each value node.
    NotConstraintComponent: URIRef  # A constraint component that can be used to verify that value nodes do not conform to a given shape.
    OrConstraintComponent: URIRef  # A constraint component that can be used to restrict the value nodes so that they conform to at least one out of several provided shapes.
    PatternConstraintComponent: URIRef  # A constraint component that can be used to verify that every value node matches a given regular expression.
    PropertyConstraintComponent: URIRef  # A constraint component that can be used to verify that all value nodes conform to the given property shape.
    QualifiedMaxCountConstraintComponent: URIRef  # A constraint component that can be used to verify that a specified maximum number of value nodes conforms to a given shape.
    QualifiedMinCountConstraintComponent: URIRef  # A constraint component that can be used to verify that a specified minimum number of value nodes conforms to a given shape.
    SPARQLConstraintComponent: URIRef  # A constraint component that can be used to define constraints based on SPARQL queries.
    UniqueLangConstraintComponent: URIRef  # A constraint component that can be used to specify that no pair of value nodes may use the same language tag.
    XoneConstraintComponent: URIRef  # A constraint component that can be used to restrict the value nodes so that they conform to exactly one out of several provided shapes.

    # http://www.w3.org/ns/shacl#NodeKind
    BlankNode: URIRef  # The node kind of all blank nodes.
    BlankNodeOrIRI: URIRef  # The node kind of all blank nodes or IRIs.
    BlankNodeOrLiteral: URIRef  # The node kind of all blank nodes or literals.
    IRI: URIRef  # The node kind of all IRIs.
    IRIOrLiteral: URIRef  # The node kind of all IRIs or literals.
    Literal: URIRef  # The node kind of all literals.

    # http://www.w3.org/ns/shacl#Parameter

    # http://www.w3.org/ns/shacl#Severity
    Info: URIRef  # The severity for an informational validation result.
    Violation: URIRef  # The severity for a violation validation result.
    Warning: URIRef  # The severity for a warning validation result.

    # Valid non-python identifiers
    _extras = [
        "and",
        "class",
        "in",
        "not",
        "or",
        "AndConstraintComponent-and",
        "ClassConstraintComponent-class",
        "ClosedConstraintComponent-closed",
        "ClosedConstraintComponent-ignoredProperties",
        "DatatypeConstraintComponent-datatype",
        "DisjointConstraintComponent-disjoint",
        "EqualsConstraintComponent-equals",
        "ExpressionConstraintComponent-expression",
        "HasValueConstraintComponent-hasValue",
        "InConstraintComponent-in",
        "JSConstraint-js",
        "LanguageInConstraintComponent-languageIn",
        "LessThanConstraintComponent-lessThan",
        "LessThanOrEqualsConstraintComponent-lessThanOrEquals",
        "MaxCountConstraintComponent-maxCount",
        "MaxExclusiveConstraintComponent-maxExclusive",
        "MaxInclusiveConstraintComponent-maxInclusive",
        "MaxLengthConstraintComponent-maxLength",
        "MinCountConstraintComponent-minCount",
        "MinExclusiveConstraintComponent-minExclusive",
        "MinInclusiveConstraintComponent-minInclusive",
        "MinLengthConstraintComponent-minLength",
        "NodeConstraintComponent-node",
        "NodeKindConstraintComponent-nodeKind",
        "NotConstraintComponent-not",
        "OrConstraintComponent-or",
        "PatternConstraintComponent-flags",
        "PatternConstraintComponent-pattern",
        "PropertyConstraintComponent-property",
        "QualifiedMaxCountConstraintComponent-qualifiedMaxCount",
        "QualifiedMaxCountConstraintComponent-qualifiedValueShape",
        "QualifiedMaxCountConstraintComponent-qualifiedValueShapesDisjoint",
        "QualifiedMinCountConstraintComponent-qualifiedMinCount",
        "QualifiedMinCountConstraintComponent-qualifiedValueShape",
        "QualifiedMinCountConstraintComponent-qualifiedValueShapesDisjoint",
        "SPARQLConstraintComponent-sparql",
        "UniqueLangConstraintComponent-uniqueLang",
        "XoneConstraintComponent-xone",
    ]

    _NS = Namespace("http://www.w3.org/ns/shacl#")



from typing import Any

from rdflib import Namespace
from rdflib.plugins.sparql.evaluate import evalPart
from rdflib.plugins.sparql.evalutils import _eval
from rdflib.plugins.sparql.sparql import SPARQLError


def custom_eval_extended(ctx: Any, extend: Any) -> Any:
    for c in evalPart(ctx, extend.p):
        try:
            if hasattr(extend.expr, "iri") and extend.expr.iri == function_uri:
                evaluation = function_result
            else:
                evaluation = _eval(extend.expr, c.forget(ctx, _except=extend._vars))
                if isinstance(evaluation, SPARQLError):
                    raise evaluation

            yield c.merge({extend.var: evaluation})

        except SPARQLError:
            yield c


def custom_eval(ctx: Any, part: Any) -> Any:
    if part.name == "Extend":
        return custom_eval_extended(ctx, part)
    else:
        raise NotImplementedError()


namespace = Namespace("example:rdflib:plugin:sparqleval:")
function_uri = namespace["function"]
function_result = namespace["result"]



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Callable, Optional, Set, Tuple
from urllib.error import HTTPError, URLError

import pytest

from rdflib import Graph, URIRef
from rdflib.exceptions import ParserError
from rdflib.namespace import Namespace, NamespaceManager
from rdflib.plugin import PluginException
from rdflib.store import Store
from rdflib.term import BNode
from test.data import BOB, CHEESE, HATES, LIKES, MICHEL, PIZZA, TAREK, TEST_DATA_DIR
from test.utils import GraphHelper, get_unique_plugin_names
from test.utils.httpfileserver import HTTPFileServer, ProtoFileResource
from test.utils.outcome import ExceptionChecker, OutcomeChecker, OutcomePrimitive


def test_property_store() -> None:
    """
    The ``store`` property works correctly.
    """
    graph = Graph()
    assert isinstance(graph.store, Store)


def test_property_identifier_default() -> None:
    """
    The default identifier for a graph is a `rdflib.term.BNode`.
    """
    graph = Graph()
    assert isinstance(graph.identifier, BNode)


def test_property_identifier() -> None:
    """
    The ``identifier`` property works correctly.
    """
    id = URIRef("example:a")
    graph = Graph(identifier=id)
    assert id == graph.identifier


def test_property_namespace_manager() -> None:
    """
    The ``namespace_manager`` property works correctly.
    """
    graph = Graph()
    # check repeats as property is a signleton
    assert isinstance(graph.namespace_manager, NamespaceManager)
    assert isinstance(graph.namespace_manager, NamespaceManager)

    new_nsm = NamespaceManager(graph)
    new_nsm.reset()
    new_nsm.bind("test", URIRef("example:test:"))
    graph.namespace_manager = new_nsm
    assert isinstance(graph.namespace_manager, NamespaceManager)
    nss = list(graph.namespace_manager.namespaces())
    assert ("test", URIRef("example:test:")) in nss


def get_store_names() -> Set[Optional[str]]:
    names: Set[Optional[str]] = {*get_unique_plugin_names(Store)}
    names.difference_update(
        {
            "default",
            "Memory",
            "Auditable",
            "Concurrent",
            "SPARQLStore",
            "SPARQLUpdateStore",
            "SimpleMemory",
        }
    )
    names.add(None)

    logging.debug("names = %s", names)
    return names


GraphFactory = Callable[[], Graph]


@pytest.fixture(scope="function", params=get_store_names())
def make_graph(tmp_path: Path, request) -> GraphFactory:
    store_name: Optional[str] = request.param

    def make_graph() -> Graph:
        if store_name is None:
            graph = Graph()
        else:
            graph = Graph(store=store_name)

        use_path = tmp_path / f"{store_name}"
        use_path.mkdir(exist_ok=True, parents=True)
        logging.debug("use_path = %s", use_path)
        graph.open(f"{use_path}", create=True)
        return graph

    return make_graph


def populate_graph(graph: Graph):
    graph.add((TAREK, LIKES, PIZZA))
    graph.add((TAREK, LIKES, CHEESE))
    graph.add((MICHEL, LIKES, PIZZA))
    graph.add((MICHEL, LIKES, CHEESE))
    graph.add((BOB, LIKES, CHEESE))
    graph.add((BOB, HATES, PIZZA))
    graph.add((BOB, HATES, MICHEL))  # gasp!


def depopulate_graph(graph: Graph):
    graph.remove((TAREK, LIKES, PIZZA))
    graph.remove((TAREK, LIKES, CHEESE))
    graph.remove((MICHEL, LIKES, PIZZA))
    graph.remove((MICHEL, LIKES, CHEESE))
    graph.remove((BOB, LIKES, CHEESE))
    graph.remove((BOB, HATES, PIZZA))
    graph.remove((BOB, HATES, MICHEL))  # gasp!


def test_add(make_graph: GraphFactory):
    graph = make_graph()
    populate_graph(graph)


def test_remove(make_graph: GraphFactory):
    graph = make_graph()
    populate_graph(graph)
    depopulate_graph(graph)


def test_triples(make_graph: GraphFactory):
    graph = make_graph()
    triples = graph.triples
    Any = None  # noqa: N806

    populate_graph(graph)

    # unbound subjects
    assert len(list(triples((Any, LIKES, PIZZA)))) == 2
    assert len(list(triples((Any, HATES, PIZZA)))) == 1
    assert len(list(triples((Any, LIKES, CHEESE)))) == 3
    assert len(list(triples((Any, HATES, CHEESE)))) == 0

    # unbound objects
    assert len(list(triples((MICHEL, LIKES, Any)))) == 2
    assert len(list(triples((TAREK, LIKES, Any)))) == 2
    assert len(list(triples((BOB, HATES, Any)))) == 2
    assert len(list(triples((BOB, LIKES, Any)))) == 1

    # unbound predicates
    assert len(list(triples((MICHEL, Any, CHEESE)))) == 1
    assert len(list(triples((TAREK, Any, CHEESE)))) == 1
    assert len(list(triples((BOB, Any, PIZZA)))) == 1
    assert len(list(triples((BOB, Any, MICHEL)))) == 1

    # unbound subject, objects
    assert len(list(triples((Any, HATES, Any)))) == 2
    assert len(list(triples((Any, LIKES, Any)))) == 5

    # unbound predicates, objects
    assert len(list(triples((MICHEL, Any, Any)))) == 2
    assert len(list(triples((BOB, Any, Any)))) == 3
    assert len(list(triples((TAREK, Any, Any)))) == 2

    # unbound subjects, predicates
    assert len(list(triples((Any, Any, PIZZA)))) == 3
    assert len(list(triples((Any, Any, CHEESE)))) == 3
    assert len(list(triples((Any, Any, MICHEL)))) == 1

    # all unbound
    assert len(list(triples((Any, Any, Any)))) == 7
    depopulate_graph(graph)
    assert len(list(triples((Any, Any, Any)))) == 0


def test_connected(make_graph: GraphFactory):
    graph = make_graph()
    populate_graph(graph)
    assert graph.connected() is True

    jeroen = URIRef("jeroen")
    unconnected = URIRef("unconnected")

    graph.add((jeroen, LIKES, unconnected))

    assert graph.connected() is False


def test_graph_sub(make_graph: GraphFactory):
    g1 = make_graph()
    g2 = make_graph()

    g1.add((TAREK, LIKES, PIZZA))
    g1.add((BOB, LIKES, CHEESE))

    g2.add((BOB, LIKES, CHEESE))

    g3 = g1 - g2

    assert len(g3) == 1
    assert (TAREK, LIKES, PIZZA) in g3
    assert (TAREK, LIKES, CHEESE) not in g3

    assert (BOB, LIKES, CHEESE) not in g3

    g1 -= g2

    assert len(g1) == 1
    assert (TAREK, LIKES, PIZZA) in g1
    assert (TAREK, LIKES, CHEESE) not in g1

    assert (BOB, LIKES, CHEESE) not in g1


def test_graph_add(make_graph: GraphFactory):
    g1 = make_graph()
    g2 = make_graph()

    g1.add((TAREK, LIKES, PIZZA))
    g2.add((BOB, LIKES, CHEESE))

    g3 = g1 + g2

    assert len(g3) == 2
    assert (TAREK, LIKES, PIZZA) in g3
    assert (TAREK, LIKES, CHEESE) not in g3

    assert (BOB, LIKES, CHEESE) in g3

    g1 += g2

    assert len(g1) == 2
    assert (TAREK, LIKES, PIZZA) in g1
    assert (TAREK, LIKES, CHEESE) not in g1

    assert (BOB, LIKES, CHEESE) in g1


def test_graph_intersection(make_graph: GraphFactory):
    g1 = make_graph()
    g2 = make_graph()

    g1.add((TAREK, LIKES, PIZZA))
    g1.add((MICHEL, LIKES, CHEESE))

    g2.add((BOB, LIKES, CHEESE))
    g2.add((MICHEL, LIKES, CHEESE))

    g3 = g1 * g2

    assert len(g3) == 1
    assert (TAREK, LIKES, PIZZA) not in g3
    assert (TAREK, LIKES, CHEESE) not in g3

    assert (BOB, LIKES, CHEESE) not in g3

    assert (MICHEL, LIKES, CHEESE) in g3

    g1 *= g2

    assert len(g1) == 1

    assert (TAREK, LIKES, PIZZA) not in g1
    assert (TAREK, LIKES, CHEESE) not in g1

    assert (BOB, LIKES, CHEESE) not in g1

    assert (MICHEL, LIKES, CHEESE) in g1


def test_guess_format_for_parse(
    make_graph: GraphFactory, http_file_server: HTTPFileServer
):
    graph = make_graph()

    # files
    with pytest.raises(ParserError):
        graph.parse(__file__)  # here we are trying to parse a Python file!!

    # .nt can be parsed by Turtle Parser
    graph.parse(os.path.join(TEST_DATA_DIR, "suites", "nt_misc", "anons-01.nt"))
    # RDF/XML
    graph.parse(
        os.path.join(
            TEST_DATA_DIR, "suites", "w3c", "rdf-xml", "datatypes", "test001.rdf"
        )
    )  # XML
    # bad filename but set format
    graph.parse(
        os.path.join(TEST_DATA_DIR, "w3c-rdfxml-test001.borked"),
        format="xml",
    )

    with pytest.raises(ParserError):
        graph.parse(data="rubbish")

    # Turtle - default
    graph.parse(
        data="<http://example.com/a> <http://example.com/a> <http://example.com/a> ."
    )

    # Turtle - format given
    graph.parse(
        data="<http://example.com/a> <http://example.com/a> <http://example.com/a> .",
        format="turtle",
    )

    # RDF/XML - format given
    rdf = """<rdf:RDF
  xmlns:ns1="http://example.org/#"
  xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
>
  <rdf:Description rdf:nodeID="ub63bL2C1">
    <ns1:p rdf:resource="http://example.org/q"/>
    <ns1:r rdf:resource="http://example.org/s"/>
  </rdf:Description>
  <rdf:Description rdf:nodeID="ub63bL5C1">
    <ns1:r>
      <rdf:Description rdf:nodeID="ub63bL6C11">
        <ns1:s rdf:resource="http://example.org/#t"/>
      </rdf:Description>
    </ns1:r>
    <ns1:p rdf:resource="http://example.org/q"/>
  </rdf:Description>
</rdf:RDF>
    """
    graph.parse(data=rdf, format="xml")

    # URI
    file_info = http_file_server.add_file_with_caching(
        ProtoFileResource(
            (("Content-Type", "text/html; charset=UTF-8"),),
            TEST_DATA_DIR / "html5lib_tests1.html",
        ),
    )

    # only getting HTML
    with pytest.raises(PluginException):
        graph.parse(location=file_info.request_url)
    try:
        # persistent Australian Government online RDF resource without a file-like ending
        graph.parse(location="https://linked.data.gov.au/def/agrif?_format=text/turtle")
    except (URLError, HTTPError):
        # this endpoint is currently not available, ignore this test.
        pass


@pytest.mark.parametrize(
    ("file", "content_type", "expected_result"),
    (
        (TEST_DATA_DIR / "defined_namespaces/adms.rdf", "application/rdf+xml", 132),
        (TEST_DATA_DIR / "defined_namespaces/adms.ttl", "text/turtle", 132),
        (TEST_DATA_DIR / "defined_namespaces/adms.ttl", None, 132),
        (
            TEST_DATA_DIR / "defined_namespaces/adms.rdf",
            None,
            ExceptionChecker(
                ParserError,
                r"Could not guess RDF format .* from file extension so tried Turtle",
            ),
        ),
    ),
)
def test_guess_format_for_parse_http(
    make_graph: GraphFactory,
    http_file_server: HTTPFileServer,
    file: Path,
    content_type: Optional[str],
    expected_result: OutcomePrimitive[int],
) -> None:
    graph = make_graph()
    headers: Tuple[Tuple[str, str], ...] = tuple()
    if content_type is not None:
        headers = (("Content-Type", content_type),)

    file_info = http_file_server.add_file_with_caching(
        ProtoFileResource(headers, file),
        suffix=f"/{file.name}",
    )
    checker = OutcomeChecker.from_primitive(expected_result)
    assert 0 == len(graph)
    with checker.context():
        graph.parse(location=file_info.request_url)
        checker.check(len(graph))


@pytest.mark.webtest
def test_guess_format_for_parse_http_text_plain():
    # Any raw url of a file from GitHub will return the content-type with text/plain.
    url = "https://raw.githubusercontent.com/AGLDWG/vocpub-profile/master/validators/validator.ttl"
    graph = Graph().parse(url)
    assert len(graph) > 0

    # A url that returns content-type text/html.
    url = "https://github.com/RDFLib/rdflib/issues/2734"
    with pytest.raises(PluginException):
        graph = Graph().parse(url)


def test_parse_file_uri(make_graph: GraphFactory):
    EG = Namespace("http://example.org/#")  # noqa: N806
    g = make_graph()
    g.parse(
        Path(os.path.join(TEST_DATA_DIR, "suites", "nt_misc", "simple-04.nt"))
        .absolute()
        .as_uri()
    )
    triple_set = GraphHelper.triple_set(g)
    assert triple_set == {
        (EG["Subject"], EG["predicate"], EG["ObjectP"]),
        (EG["Subject"], EG["predicate"], EG["ObjectQ"]),
        (EG["Subject"], EG["predicate"], EG["ObjectR"]),
    }


def test_transitive(make_graph: GraphFactory):
    person = URIRef("ex:person")
    dad = URIRef("ex:dad")
    mom = URIRef("ex:mom")
    mom_of_dad = URIRef("ex:mom_o_dad")
    mom_of_mom = URIRef("ex:mom_o_mom")
    dad_of_dad = URIRef("ex:dad_o_dad")
    dad_of_mom = URIRef("ex:dad_o_mom")

    parent = URIRef("ex:parent")

    g = make_graph()
    g.add((person, parent, dad))
    g.add((person, parent, mom))
    g.add((dad, parent, mom_of_dad))
    g.add((dad, parent, dad_of_dad))
    g.add((mom, parent, mom_of_mom))
    g.add((mom, parent, dad_of_mom))

    # transitive parents of person
    assert set(g.transitive_objects(subject=person, predicate=parent)) == {
        person,
        dad,
        mom_of_dad,
        dad_of_dad,
        mom,
        mom_of_mom,
        dad_of_mom,
    }
    # transitive parents of dad
    assert set(g.transitive_objects(dad, parent)) == {dad, mom_of_dad, dad_of_dad}
    # transitive parents of dad_of_dad
    assert set(g.transitive_objects(dad_of_dad, parent)) == {dad_of_dad}

    # transitive children (inverse of parents) of mom_of_mom
    assert set(g.transitive_subjects(predicate=parent, object=mom_of_mom)) == {
        mom_of_mom,
        mom,
        person,
    }
    # transitive children (inverse of parents) of mom
    assert set(g.transitive_subjects(parent, mom)) == {mom, person}
    # transitive children (inverse of parents) of person
    assert set(g.transitive_subjects(parent, person)) == {person}



import os
from contextlib import contextmanager
from pathlib import PurePath
from typing import Iterator, TypeVar, Union

PathLike = Union[PurePath, str]
PathLikeT = TypeVar("PathLikeT", bound=PathLike)


@contextmanager
def ctx_chdir(newdir: PathLikeT) -> Iterator[PathLikeT]:
    cwd = os.getcwd()
    try:
        os.chdir(f"{newdir}")
        yield newdir
    finally:
        os.chdir(cwd)



from __future__ import annotations

import enum
import http.client
import itertools
import logging
from contextlib import ExitStack
from pathlib import Path
from textwrap import dedent
from typing import Any, Iterable, Tuple
from urllib.request import HTTPHandler, OpenerDirector, Request

import pytest
from _pytest.mark.structures import ParameterSet

from rdflib import Graph
from test.utils.audit import AuditHookDispatcher
from test.utils.httpfileserver import HTTPFileServer, ProtoFileResource
from test.utils.namespace import EGDO
from test.utils.urlopen import context_urlopener

from ..utils import GraphHelper
from ..utils.path import ctx_chdir

JSONLD_CONTEXT = """
{
    "@context": {
        "ex": "http://example.org/"
    }
}
"""

EXPECTED_GRAPH = Graph().add((EGDO.subject, EGDO.predicate, EGDO.object))


def test_default(tmp_path: Path) -> None:
    context_file = tmp_path / "context.jsonld"
    context_file.write_text(dedent(JSONLD_CONTEXT))

    data = f"""
    {{
        "@context": "{context_file.as_uri()}",
        "@id": "ex:subject",
        "ex:predicate": {{ "@id": "ex:object" }}
    }}
    """

    graph = Graph()
    graph.parse(format="json-ld", data=data)
    logging.debug("graph = %s", GraphHelper.triple_set(graph))
    GraphHelper.assert_sets_equals(EXPECTED_GRAPH, graph)


class Defence(enum.Enum):
    NONE = enum.auto()
    AUDIT_HOOK = enum.auto()
    URL_OPENER = enum.auto()


class URIKind(enum.Enum):
    FILE = enum.auto()
    HTTP = enum.auto()
    RELATIVE = enum.auto()


def generate_make_block_file_cases() -> Iterable[ParameterSet]:
    for defence, uri_kind in itertools.product(Defence, URIKind):
        if defence == Defence.URL_OPENER and uri_kind != URIKind.HTTP:
            # URL opener only works for not file URIs
            continue
        yield pytest.param(defence, uri_kind)


@pytest.mark.parametrize(["defence", "uri_kind"], generate_make_block_file_cases())
def test_block_file(
    tmp_path: Path,
    audit_hook_dispatcher: AuditHookDispatcher,
    http_file_server: HTTPFileServer,
    exit_stack: ExitStack,
    defence: Defence,
    uri_kind: URIKind,
) -> None:
    context_file = tmp_path / "context.jsonld"
    context_file.write_text(dedent(JSONLD_CONTEXT))
    context_file_served = http_file_server.add_file_with_caching(
        ProtoFileResource((), context_file)
    )

    context_uri: str
    if uri_kind == URIKind.FILE:
        context_uri = context_file.as_uri()
    elif uri_kind == URIKind.HTTP:
        context_uri = context_file_served.request_url
    elif uri_kind == URIKind.RELATIVE:
        context_uri = context_file.name
        exit_stack.enter_context(ctx_chdir(tmp_path))
    else:
        raise ValueError(f"unknown URI kind: {uri_kind}")

    data = f"""
    {{
        "@context": "{context_uri}",
        "@id": "ex:subject",
        "ex:predicate": {{ "@id": "ex:object" }}
    }}
    """

    data_file = tmp_path / "data.jsonld"
    data_file.write_text(dedent(data))

    if defence == Defence.AUDIT_HOOK and uri_kind == URIKind.FILE:

        def audit_hook(name: str, args: Tuple[Any, ...]) -> None:
            logging.info("block_file_access: name = %s, args = %s", name, args)
            if name == "open" and args[0] == f"{context_file.absolute()}":
                raise PermissionError("access blocked")

        exit_stack.enter_context(audit_hook_dispatcher.ctx_hook("open", audit_hook))

    elif defence == Defence.AUDIT_HOOK and uri_kind == URIKind.RELATIVE:

        def audit_hook(name: str, args: Tuple[Any, ...]) -> None:
            logging.info("block_file_access: name = %s, args = %s", name, args)
            if name == "open" and args[0] == f"{Path.cwd() / context_file.name}":
                raise PermissionError("access blocked")

        exit_stack.enter_context(audit_hook_dispatcher.ctx_hook("open", audit_hook))

    elif defence == Defence.AUDIT_HOOK and uri_kind == URIKind.HTTP:

        def audit_hook(name: str, args: Tuple[Any, ...]) -> None:
            logging.info("block_file_access: name = %s, args = %s", name, args)
            if name == "urllib.Request" and args[0] == context_file_served.request_url:
                raise PermissionError("access blocked")

        exit_stack.enter_context(
            audit_hook_dispatcher.ctx_hook("urllib.Request", audit_hook)
        )

    elif defence == Defence.URL_OPENER and uri_kind == URIKind.HTTP:
        opener = OpenerDirector()

        class SecuredHTTPHandler(HTTPHandler):
            def http_open(self, req: Request) -> http.client.HTTPResponse:
                if req.get_full_url() == context_file_served.request_url:
                    raise PermissionError("access blocked")
                return super().http_open(req)

        opener.add_handler(SecuredHTTPHandler())

        exit_stack.enter_context(context_urlopener(opener))

    elif defence == Defence.NONE:
        pass
    else:
        raise ValueError(
            f"unsupported defence {defence} and uri_kind {uri_kind} combination"
        )

    graph = Graph()
    if defence != Defence.NONE:
        with pytest.raises(PermissionError):
            graph.parse(format="json-ld", data=data)
        assert len(graph) == 0
    else:
        graph.parse(format="json-ld", data=data)
        GraphHelper.assert_sets_equals(EXPECTED_GRAPH, graph)



"""
This parser will interpret a JSON-LD document as an RDF Graph. See:

    http://json-ld.org/

Example usage::

    >>> from rdflib import Graph, URIRef, Literal
    >>> test_json = '''
    ... {
    ...     "@context": {
    ...         "dc": "http://purl.org/dc/terms/",
    ...         "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    ...         "rdfs": "http://www.w3.org/2000/01/rdf-schema#"
    ...     },
    ...     "@id": "http://example.org/about",
    ...     "dc:title": {
    ...         "@language": "en",
    ...         "@value": "Someone's Homepage"
    ...     }
    ... }
    ... '''
    >>> g = Graph().parse(data=test_json, format='json-ld')
    >>> list(g) == [(URIRef('http://example.org/about'),
    ...     URIRef('http://purl.org/dc/terms/title'),
    ...     Literal("Someone's Homepage", lang='en'))]
    True

"""

# From: https://github.com/RDFLib/rdflib-jsonld/blob/feature/json-ld-1.1/rdflib_jsonld/parser.py

# NOTE: This code reads the entire JSON object into memory before parsing, but
# we should consider streaming the input to deal with arbitrarily large graphs.
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Union

import rdflib.parser
from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.namespace import RDF, XSD
from rdflib.parser import InputSource, URLInputSource
from rdflib.term import BNode, IdentifiedNode, Literal, Node, URIRef

from ..shared.jsonld.context import UNDEF, Context, Term
from ..shared.jsonld.keys import (
    CONTEXT,
    GRAPH,
    ID,
    INCLUDED,
    INDEX,
    JSON,
    LANG,
    LIST,
    NEST,
    NONE,
    REV,
    SET,
    TYPE,
    VALUE,
    VOCAB,
)
from ..shared.jsonld.util import (
    _HAS_ORJSON,
    VOCAB_DELIMS,
    context_from_urlinputsource,
    json,
    orjson,
    source_to_json,
)

__all__ = ["JsonLDParser", "to_rdf"]

TYPE_TERM = Term(str(RDF.type), TYPE, VOCAB)  # type: ignore[call-arg]

ALLOW_LISTS_OF_LISTS = True  # NOTE: Not allowed in JSON-LD 1.0


class JsonLDParser(rdflib.parser.Parser):
    def __init__(self):
        super(JsonLDParser, self).__init__()

    def parse(
        self,
        source: InputSource,
        sink: Graph,
        version: float = 1.1,
        skolemize: bool = False,
        encoding: Optional[str] = "utf-8",
        base: Optional[str] = None,
        context: Optional[
            Union[
                List[Union[Dict[str, Any], str, None]],
                Dict[str, Any],
                str,
            ]
        ] = None,
        generalized_rdf: Optional[bool] = False,
        extract_all_scripts: Optional[bool] = False,
        **kwargs: Any,
    ) -> None:
        """Parse JSON-LD from a source document.

        The source document can be JSON or HTML with embedded JSON script
        elements (type attribute = "application/ld+json"). To process as HTML
        ``source.content_type`` must be set to "text/html" or
        "application/xhtml+xml".

        :param source: InputSource with JSON-formatted data (JSON or HTML)

        :param sink: Graph to receive the parsed triples

        :param version: parse as JSON-LD version, defaults to 1.1

        :param encoding: character encoding of the JSON (should be "utf-8"
            or "utf-16"), defaults to "utf-8"

        :param base: JSON-LD `Base IRI <https://www.w3.org/TR/json-ld/#base-iri>`_, defaults to None

        :param context: JSON-LD `Context <https://www.w3.org/TR/json-ld/#the-context>`_, defaults to None

        :param generalized_rdf: parse as `Generalized RDF <https://www.w3.org/TR/json-ld/#relationship-to-rdf>`_, defaults to False

        :param extract_all_scripts: if source is an HTML document then extract
            all script elements, defaults to False (extract only the first
            script element). This is ignored if ``source.system_id`` contains
            a fragment identifier, in which case only the script element with
            matching id attribute is extracted.

        """
        if encoding not in ("utf-8", "utf-16"):
            warnings.warn(
                "JSON should be encoded as unicode. "
                "Given encoding was: %s" % encoding
            )

        if not base:
            base = sink.absolutize(source.getPublicId() or source.getSystemId() or "")

        context_data = context
        if not context_data and hasattr(source, "url") and hasattr(source, "links"):
            if TYPE_CHECKING:
                assert isinstance(source, URLInputSource)
            context_data = context_from_urlinputsource(source)

        try:
            version = float(version)
        except ValueError:
            version = 1.1

        # Get the optional fragment identifier
        try:
            fragment_id = URIRef(source.getSystemId()).fragment
        except Exception:
            fragment_id = None

        data, html_base = source_to_json(source, fragment_id, extract_all_scripts)
        if html_base is not None:
            base = URIRef(html_base, base=base)

        # NOTE: A ConjunctiveGraph parses into a Graph sink, so no sink will be
        # context_aware. Keeping this check in case RDFLib is changed, or
        # someone passes something context_aware to this parser directly.
        conj_sink: Graph
        if not sink.context_aware:
            conj_sink = ConjunctiveGraph(store=sink.store, identifier=sink.identifier)
        else:
            conj_sink = sink

        to_rdf(
            data,
            conj_sink,
            base,
            context_data,
            version,
            bool(generalized_rdf),
            skolemize=skolemize,
        )


def to_rdf(
    data: Any,
    dataset: Graph,
    base: Optional[str] = None,
    context_data: Optional[
        Union[
            List[Union[Dict[str, Any], str, None]],
            Dict[str, Any],
            str,
        ]
    ] = None,
    version: Optional[float] = None,
    generalized_rdf: bool = False,
    allow_lists_of_lists: Optional[bool] = None,
    skolemize: bool = False,
):
    # TODO: docstring w. args and return value
    context = Context(base=base, version=version)
    if context_data:
        context.load(context_data)
    parser = Parser(
        generalized_rdf=generalized_rdf,
        allow_lists_of_lists=allow_lists_of_lists,
        skolemize=skolemize,
    )
    return parser.parse(data, context, dataset)


class Parser:
    def __init__(
        self,
        generalized_rdf: bool = False,
        allow_lists_of_lists: Optional[bool] = None,
        skolemize: bool = False,
    ):
        self.skolemize = skolemize
        self.generalized_rdf = generalized_rdf
        self.allow_lists_of_lists = (
            allow_lists_of_lists
            if allow_lists_of_lists is not None
            else ALLOW_LISTS_OF_LISTS
        )

    def parse(self, data: Any, context: Context, dataset: Graph) -> Graph:
        topcontext = False
        resources: Union[Dict[str, Any], List[Any]]
        if isinstance(data, list):
            resources = data
        elif isinstance(data, dict):
            local_context = data.get(CONTEXT)
            if local_context:
                context.load(local_context, context.base)
                topcontext = True
            resources = data
            # type error: Subclass of "Dict[str, Any]" and "List[Any]" cannot exist: would have incompatible method signatures
            if not isinstance(resources, list):  # type: ignore[unreachable]
                resources = [resources]

        if context.vocab:
            dataset.bind(None, context.vocab)
        for name, term in context.terms.items():
            if term.id and term.id.endswith(VOCAB_DELIMS):
                dataset.bind(name, term.id)

        # type error: "Graph" has no attribute "default_context"
        graph = dataset.default_context if dataset.context_aware else dataset  # type: ignore[attr-defined]

        for node in resources:
            self._add_to_graph(dataset, graph, context, node, topcontext)

        return graph

    def _add_to_graph(
        self,
        dataset: Graph,
        graph: Graph,
        context: Context,
        node: Any,
        topcontext: bool = False,
    ) -> Optional[Node]:
        if not isinstance(node, dict) or context.get_value(node):
            # type error: Return value expected
            return  # type: ignore[return-value]

        if CONTEXT in node and not topcontext:
            local_context = node[CONTEXT]
            if local_context:
                context = context.subcontext(local_context)
            else:
                context = Context(base=context.doc_base)

        # type error: Incompatible types in assignment (expression has type "Optional[Context]", variable has type "Context")
        context = context.get_context_for_type(node)  # type: ignore[assignment]

        id_val = context.get_id(node)

        if id_val is None:
            nested_id = self._get_nested_id(context, node)
            if nested_id is not None and len(nested_id) > 0:
                id_val = nested_id

        if isinstance(id_val, str):
            subj = self._to_rdf_id(context, id_val)
        else:
            subj = BNode()
            if self.skolemize:
                subj = subj.skolemize()

        if subj is None:
            return None

        # NOTE: crude way to signify that this node might represent a named graph
        no_id = id_val is None

        for key, obj in node.items():
            if key == CONTEXT or key in context.get_keys(ID):
                continue

            if key == REV or key in context.get_keys(REV):
                for rkey, robj in obj.items():
                    self._key_to_graph(
                        dataset,
                        graph,
                        context,
                        subj,
                        rkey,
                        robj,
                        reverse=True,
                        no_id=no_id,
                    )
            else:
                self._key_to_graph(dataset, graph, context, subj, key, obj, no_id=no_id)

        return subj

    # type error: Missing return statement
    def _get_nested_id(self, context: Context, node: Dict[str, Any]) -> Optional[str]:  # type: ignore[return]
        for key, obj in node.items():
            if context.version >= 1.1 and key in context.get_keys(NEST):
                term = context.terms.get(key)
                if term and term.id is None:
                    continue
                objs = obj if isinstance(obj, list) else [obj]
                for obj in objs:
                    if not isinstance(obj, dict):
                        continue
                    id_val = context.get_id(obj)
                    if not id_val:
                        subcontext = context.get_context_for_term(
                            context.terms.get(key)
                        )
                        id_val = self._get_nested_id(subcontext, obj)
                    if isinstance(id_val, str):
                        return id_val

    def _key_to_graph(
        self,
        dataset: Graph,
        graph: Graph,
        context: Context,
        subj: Node,
        key: str,
        obj: Any,
        reverse: bool = False,
        no_id: bool = False,
    ) -> None:
        if isinstance(obj, list):
            obj_nodes = obj
        else:
            obj_nodes = [obj]

        term = context.terms.get(key)
        if term:
            term_id = term.id
            if term.type == JSON:
                obj_nodes = [self._to_typed_json_value(obj)]
            elif LIST in term.container:
                obj_nodes = [self._expand_nested_list(obj_nodes)]
            elif isinstance(obj, dict):
                obj_nodes = self._parse_container(context, term, obj)
        else:
            term_id = None

        if TYPE in (key, term_id):
            term = TYPE_TERM

        if GRAPH in (key, term_id):
            if dataset.context_aware and not no_id:
                if TYPE_CHECKING:
                    assert isinstance(dataset, ConjunctiveGraph)
                # type error: Argument 1 to "get_context" of "ConjunctiveGraph" has incompatible type "Node"; expected "Union[IdentifiedNode, str, None]"
                subgraph = dataset.get_context(subj)  # type: ignore[arg-type]
            else:
                subgraph = graph
            for onode in obj_nodes:
                self._add_to_graph(dataset, subgraph, context, onode)
            return

        if SET in (key, term_id):
            for onode in obj_nodes:
                self._add_to_graph(dataset, graph, context, onode)
            return

        if INCLUDED in (key, term_id):
            for onode in obj_nodes:
                self._add_to_graph(dataset, graph, context, onode)
            return

        if context.version >= 1.1 and key in context.get_keys(NEST):
            term = context.terms.get(key)
            if term and term.id is None:
                return
            objs = obj if isinstance(obj, list) else [obj]
            for obj in objs:
                if not isinstance(obj, dict):
                    continue
                for nkey, nobj in obj.items():
                    # NOTE: we've already captured subject
                    if nkey in context.get_keys(ID):
                        continue
                    subcontext = context.get_context_for_type(obj)
                    # type error: Argument 3 to "_key_to_graph" of "Parser" has incompatible type "Optional[Context]"; expected "Context"
                    self._key_to_graph(dataset, graph, subcontext, subj, nkey, nobj)  # type: ignore[arg-type]
            return

        pred_uri = term.id if term else context.expand(key)

        context = context.get_context_for_term(term)

        # Flatten deep nested lists
        def flatten(n: Iterable[Any]) -> List[Any]:
            flattened = []
            for obj in n:
                if isinstance(obj, dict):
                    objs = context.get_set(obj)
                    if objs is not None:
                        obj = objs
                if isinstance(obj, list):
                    flattened += flatten(obj)
                    continue
                flattened.append(obj)
            return flattened

        obj_nodes = flatten(obj_nodes)

        if not pred_uri:
            return

        if term and term.reverse:
            reverse = not reverse

        pred: IdentifiedNode
        bid = self._get_bnodeid(pred_uri)
        if bid:
            if not self.generalized_rdf:
                return
            pred = BNode(bid)
            if self.skolemize:
                pred = pred.skolemize()
        else:
            pred = URIRef(pred_uri)

        for obj_node in obj_nodes:
            obj = self._to_object(dataset, graph, context, term, obj_node)
            if obj is None:
                continue
            if reverse:
                graph.add((obj, pred, subj))
            else:
                graph.add((subj, pred, obj))

    def _parse_container(
        self, context: Context, term: Term, obj: Dict[str, Any]
    ) -> List[Any]:
        if LANG in term.container:
            obj_nodes = []
            for lang, values in obj.items():
                if not isinstance(values, list):
                    values = [values]
                if lang in context.get_keys(NONE):
                    obj_nodes += values
                else:
                    for v in values:
                        obj_nodes.append((v, lang))
            return obj_nodes

        v11 = context.version >= 1.1

        if v11 and GRAPH in term.container and ID in term.container:
            return [
                (
                    dict({GRAPH: o})
                    if k in context.get_keys(NONE)
                    else dict({ID: k, GRAPH: o}) if isinstance(o, dict) else o
                )
                for k, o in obj.items()
            ]

        elif v11 and GRAPH in term.container and INDEX in term.container:
            return [dict({GRAPH: o}) for k, o in obj.items()]

        elif v11 and GRAPH in term.container:
            return [dict({GRAPH: obj})]

        elif v11 and ID in term.container:
            return [
                (
                    dict({ID: k}, **o)
                    if isinstance(o, dict) and k not in context.get_keys(NONE)
                    else o
                )
                for k, o in obj.items()
            ]

        elif v11 and TYPE in term.container:
            return [
                (
                    self._add_type(
                        context,
                        (
                            {ID: context.expand(o) if term.type == VOCAB else o}
                            if isinstance(o, str)
                            else o
                        ),
                        k,
                    )
                    if isinstance(o, (dict, str)) and k not in context.get_keys(NONE)
                    else o
                )
                for k, o in obj.items()
            ]

        elif INDEX in term.container:
            obj_nodes = []
            for key, nodes in obj.items():
                if not isinstance(nodes, list):
                    nodes = [nodes]
                for node in nodes:
                    if v11 and term.index and key not in context.get_keys(NONE):
                        if not isinstance(node, dict):
                            node = {ID: node}
                        values = node.get(term.index, [])
                        if not isinstance(values, list):
                            values = [values]
                        values.append(key)
                        node[term.index] = values
                    obj_nodes.append(node)
            return obj_nodes

        return [obj]

    @staticmethod
    def _add_type(context: Context, o: Dict[str, Any], k: str) -> Dict[str, Any]:
        otype = context.get_type(o) or []
        if otype and not isinstance(otype, list):
            otype = [otype]
        otype.append(k)
        o[TYPE] = otype
        return o

    def _to_object(
        self,
        dataset: Graph,
        graph: Graph,
        context: Context,
        term: Optional[Term],
        node: Any,
        inlist: bool = False,
    ) -> Optional[Node]:
        if isinstance(node, tuple):
            value, lang = node
            if value is None:
                # type error: Return value expected
                return  # type: ignore[return-value]
            if lang and " " in lang:
                # type error: Return value expected
                return  # type: ignore[return-value]
            return Literal(value, lang=lang)

        if isinstance(node, dict):
            node_list = context.get_list(node)
            if node_list is not None:
                if inlist and not self.allow_lists_of_lists:
                    # type error: Return value expected
                    return  # type: ignore[return-value]
                listref = self._add_list(dataset, graph, context, term, node_list)
                if listref:
                    return listref

        else:  # expand compacted value
            if term and term.type:
                if term.type == JSON:
                    node = self._to_typed_json_value(node)
                elif node is None:
                    # type error: Return value expected
                    return  # type: ignore[return-value]
                elif term.type == ID and isinstance(node, str):
                    node = {ID: context.resolve(node)}
                elif term.type == VOCAB and isinstance(node, str):
                    node = {ID: context.expand(node) or context.resolve_iri(node)}
                else:
                    node = {TYPE: term.type, VALUE: node}
            else:
                if node is None:
                    # type error: Return value expected
                    return  # type: ignore[return-value]
                if isinstance(node, float):
                    return Literal(node, datatype=XSD.double)

                if term and term.language is not UNDEF:
                    lang = term.language
                else:
                    lang = context.language
                return Literal(node, lang=lang)

        lang = context.get_language(node)
        datatype = not lang and context.get_type(node) or None
        value = context.get_value(node)
        # type error: Unsupported operand types for in ("Optional[Any]" and "Generator[str, None, None]")
        if datatype in context.get_keys(JSON):  # type: ignore[operator]
            node = self._to_typed_json_value(value)
            datatype = context.get_type(node)
            value = context.get_value(node)

        if lang or context.get_key(VALUE) in node or VALUE in node:
            if value is None:
                return None
            if lang:
                if " " in lang:
                    # type error: Return value expected
                    return  # type: ignore[return-value]
                return Literal(value, lang=lang)
            elif datatype:
                return Literal(value, datatype=context.expand(datatype))
            else:
                return Literal(value)
        else:
            return self._add_to_graph(dataset, graph, context, node)

    def _to_rdf_id(self, context: Context, id_val: str) -> Optional[IdentifiedNode]:
        bid = self._get_bnodeid(id_val)
        if bid:
            b = BNode(bid)
            if self.skolemize:
                return b.skolemize()
            return b
        else:
            uri = context.resolve(id_val)
            if not self.generalized_rdf and ":" not in uri:
                return None
            return URIRef(uri)

    def _get_bnodeid(self, ref: str) -> Optional[str]:
        if not ref.startswith("_:"):
            # type error: Return value expected
            return  # type: ignore[return-value]
        bid = ref.split("_:", 1)[-1]
        return bid or None

    def _add_list(
        self,
        dataset: Graph,
        graph: Graph,
        context: Context,
        term: Optional[Term],
        node_list: Any,
    ) -> IdentifiedNode:
        if not isinstance(node_list, list):
            node_list = [node_list]

        first_subj: Union[URIRef, BNode] = BNode()
        if self.skolemize and isinstance(first_subj, BNode):
            first_subj = first_subj.skolemize()

        rest: Union[URIRef, BNode, None]
        subj, rest = first_subj, None

        for node in node_list:
            if node is None:
                continue

            if rest:
                # type error: Statement is unreachable
                graph.add((subj, RDF.rest, rest))  # type: ignore[unreachable]
                subj = rest

            obj = self._to_object(dataset, graph, context, term, node, inlist=True)

            if obj is None:
                continue

            graph.add((subj, RDF.first, obj))
            rest = BNode()
            if self.skolemize and isinstance(rest, BNode):
                rest = rest.skolemize()

        if rest:
            graph.add((subj, RDF.rest, RDF.nil))
            return first_subj
        else:
            return RDF.nil

    @staticmethod
    def _to_typed_json_value(value: Any) -> Dict[str, str]:
        if _HAS_ORJSON:
            val_string: str = orjson.dumps(
                value,
                option=orjson.OPT_SORT_KEYS | orjson.OPT_NON_STR_KEYS,
            ).decode("utf-8")
        else:
            val_string = json.dumps(
                value, separators=(",", ":"), sort_keys=True, ensure_ascii=False
            )
        return {
            TYPE: RDF.JSON,
            VALUE: val_string,
        }

    @classmethod
    def _expand_nested_list(cls, obj_nodes: List[Any]) -> Dict[str, List[Any]]:
        result = [
            cls._expand_nested_list(o) if isinstance(o, list) else o for o in obj_nodes
        ]
        return {LIST: result}



import warnings
from random import randint

from rdflib.namespace import RDF
from rdflib.term import BNode, URIRef

__all__ = ["Container", "Bag", "Seq", "Alt", "NoElementException"]


class Container:
    """A class for constructing RDF containers, as per https://www.w3.org/TR/rdf11-mt/#rdf-containers

    Basic usage, creating a ``Bag`` and adding to it::

        >>> from rdflib import Graph, BNode, Literal, Bag
        >>> g = Graph()
        >>> b = Bag(g, BNode(), [Literal("One"), Literal("Two"), Literal("Three")])
        >>> print(g.serialize(format="turtle"))
        @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
        <BLANKLINE>
        [] a rdf:Bag ;
            rdf:_1 "One" ;
            rdf:_2 "Two" ;
            rdf:_3 "Three" .
        <BLANKLINE>
        <BLANKLINE>

        >>> # print out an item using an index reference
        >>> print(b[2])
        Two

        >>> # add a new item
        >>> b.append(Literal("Hello")) # doctest: +ELLIPSIS
        <rdflib.container.Bag object at ...>
        >>> print(g.serialize(format="turtle"))
        @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
        <BLANKLINE>
        [] a rdf:Bag ;
            rdf:_1 "One" ;
            rdf:_2 "Two" ;
            rdf:_3 "Three" ;
            rdf:_4 "Hello" .
        <BLANKLINE>
        <BLANKLINE>

    """

    def __init__(self, graph, uri, seq=[], rtype="Bag"):
        """Creates a Container

        :param graph: a Graph instance
        :param uri: URI or Blank Node of the Container
        :param seq: the elements of the Container
        :param rtype: the type of Container, one of "Bag", "Seq" or "Alt"
        """

        self.graph = graph
        self.uri = uri or BNode()
        self._len = 0
        self._rtype = rtype  # rdf:Bag or rdf:Seq or rdf:Alt

        self.append_multiple(seq)

        # adding triple corresponding to container type
        self.graph.add((self.uri, RDF.type, RDF[self._rtype]))

    def n3(self):
        items = []
        for i in range(len(self)):
            v = self[i + 1]
            items.append(v)

        return "( %s )" % " ".join([a.n3() for a in items])

    def _get_container(self):
        """Returns the URI of the container"""

        return self.uri

    def __len__(self):
        """Number of items in container"""

        return self._len

    def type_of_conatiner(self):
        warnings.warn(
            "rdflib.container.Container.type_of_conatiner is deprecated. "
            "Use type_of_container method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._rtype

    def type_of_container(self):
        return self._rtype

    def index(self, item):
        """Returns the 1-based numerical index of the item in the container"""

        pred = self.graph.predicates(self.uri, item)
        if not pred:
            raise ValueError("%s is not in %s" % (item, "container"))
        li_index = URIRef(str(RDF) + "_")

        i = None
        for p in pred:
            i = int(p.replace(li_index, ""))
        return i

    def __getitem__(self, key):
        """Returns item of the container at index key"""

        c = self._get_container()

        assert isinstance(key, int)
        elem_uri = str(RDF) + "_" + str(key)
        if key <= 0 or key > len(self):
            raise KeyError(key)
        v = self.graph.value(c, URIRef(elem_uri))
        if v:
            return v
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        """Sets the item at index key or predicate rdf:_key of the container to value"""

        assert isinstance(key, int)

        c = self._get_container()
        elem_uri = str(RDF) + "_" + str(key)
        if key <= 0 or key > len(self):
            raise KeyError(key)

        self.graph.set((c, URIRef(elem_uri), value))

    def __delitem__(self, key):
        """Removing the item with index key or predicate rdf:_key"""

        assert isinstance(key, int)
        if key <= 0 or key > len(self):
            raise KeyError(key)

        graph = self.graph
        container = self.uri
        elem_uri = str(RDF) + "_" + str(key)
        graph.remove((container, URIRef(elem_uri), None))
        for j in range(key + 1, len(self) + 1):
            elem_uri = str(RDF) + "_" + str(j)
            v = graph.value(container, URIRef(elem_uri))
            graph.remove((container, URIRef(elem_uri), v))
            elem_uri = str(RDF) + "_" + str(j - 1)
            graph.add((container, URIRef(elem_uri), v))

        self._len -= 1

    def items(self):
        """Returns a list of all items in the container"""

        l_ = []
        container = self.uri
        i = 1
        while True:
            elem_uri = str(RDF) + "_" + str(i)

            if (container, URIRef(elem_uri), None) in self.graph:
                i += 1
                l_.append(self.graph.value(container, URIRef(elem_uri)))
            else:
                break
        return l_

    def end(self):  #
        # find end index (1-based) of container

        container = self.uri
        i = 1
        while True:
            elem_uri = str(RDF) + "_" + str(i)

            if (container, URIRef(elem_uri), None) in self.graph:
                i += 1
            else:
                return i - 1

    def append(self, item):
        """Adding item to the end of the container"""

        end = self.end()
        elem_uri = str(RDF) + "_" + str(end + 1)
        container = self.uri
        self.graph.add((container, URIRef(elem_uri), item))
        self._len += 1

        return self

    def append_multiple(self, other):
        """Adding multiple elements to the container to the end which are in python list other"""

        end = self.end()  # it should return the last index

        container = self.uri
        for item in other:
            end += 1
            self._len += 1
            elem_uri = str(RDF) + "_" + str(end)
            self.graph.add((container, URIRef(elem_uri), item))

        return self

    def clear(self):
        """Removing all elements from the container"""

        container = self.uri
        graph = self.graph
        i = 1
        while True:
            elem_uri = str(RDF) + "_" + str(i)
            if (container, URIRef(elem_uri), None) in self.graph:
                graph.remove((container, URIRef(elem_uri), None))
                i += 1
            else:
                break
        self._len = 0
        return self


class Bag(Container):
    """Unordered container (no preference order of elements)"""

    def __init__(self, graph, uri, seq=[]):
        Container.__init__(self, graph, uri, seq, "Bag")


class Alt(Container):
    def __init__(self, graph, uri, seq=[]):
        Container.__init__(self, graph, uri, seq, "Alt")

    def anyone(self):
        if len(self) == 0:
            raise NoElementException()
        else:
            p = randint(1, len(self))
            item = self.__getitem__(p)
            return item


class Seq(Container):
    def __init__(self, graph, uri, seq=[]):
        Container.__init__(self, graph, uri, seq, "Seq")

    def add_at_position(self, pos, item):
        assert isinstance(pos, int)
        if pos <= 0 or pos > len(self) + 1:
            raise ValueError("Invalid Position for inserting element in rdf:Seq")

        if pos == len(self) + 1:
            self.append(item)
        else:
            for j in range(len(self), pos - 1, -1):
                container = self._get_container()
                elem_uri = str(RDF) + "_" + str(j)
                v = self.graph.value(container, URIRef(elem_uri))
                self.graph.remove((container, URIRef(elem_uri), v))
                elem_uri = str(RDF) + "_" + str(j + 1)
                self.graph.add((container, URIRef(elem_uri), v))
            elem_uri_pos = str(RDF) + "_" + str(pos)
            self.graph.add((container, URIRef(elem_uri_pos), item))
            self._len += 1
        return self


class NoElementException(Exception):  # noqa: N818
    def __init__(self, message="rdf:Alt Container is empty"):
        self.message = message

    def __str__(self):
        return self.message



import itertools
import os
from urllib.error import URLError

import pytest

from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.plugins.parsers.notation3 import BadSyntax, exponent_syntax
from rdflib.term import Literal, URIRef
from test import TEST_DIR

test_data = """
#  Definitions of terms describing the n3 model
#

@keywords a.

@prefix n3: <#>.
@prefix log: <log.n3#> .
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <#> .

@forAll :s, :p, :x, :y, :z.

n3:Statement    a rdf:Class .
n3:StatementSet a rdf:Class .

n3:includes     a rdfs:Property .   # Cf rdf:li

n3:predicate    a rdf:Property; rdfs:domain n3:statement .
n3:subject      a rdf:Property; rdfs:domain n3:statement .
n3:object       a rdf:Property; rdfs:domain n3:statement .

n3:context      a rdf:Property; rdfs:domain n3:statement;
                rdfs:range n3:StatementSet .



########### Rules

{ :x :p :y . } log:means { [
                n3:subject :x;
                n3:predicate :p;
                n3:object :y ] a log:Truth}.

# Needs more thought ... ideally, we have the implcit AND rules of
# juxtaposition (introduction and elimination)

{
    {
        {  :x n3:includes :s. } log:implies { :y n3:includes :s. } .
    } forall :s1 .
} log:implies { :x log:implies :y } .

{
    {
        {  :x n3:includes :s. } log:implies { :y n3:includes :s. } .
    } forall :s1
} log:implies { :x log:implies :y } .

# I think n3:includes has to be axiomatic builtin. - unless you go to syntax description.
# syntax.n3?
"""


class TestN3Case:
    def test_base_cumulative(self):
        """
        Test that the n3 parser supports base declarations
        This is issue #22
        """

        input = """
@prefix : <http://example.com/> .
# default base
<foo> :name "Foo" .
# change it
@base <http://example.com/doc/> .
<bar> :name "Bar" .
# and change it more - they are cumulative
@base <doc2/> .
<bing> :name "Bing" .
# unless abosulute
@base <http://test.com/> .
<bong> :name "Bong" .

"""
        g = Graph()
        g.parse(data=input, format="n3")
        print(list(g))
        assert (None, None, Literal("Foo")) in g
        assert (URIRef("http://example.com/doc/bar"), None, None) in g
        assert (URIRef("http://example.com/doc/doc2/bing"), None, None) in g
        assert (URIRef("http://test.com/bong"), None, None) in g

    def test_base_explicit(self):
        """
        Test that the n3 parser supports resolving relative URIs
        and that base will override
        """

        input = """
@prefix : <http://example.com/> .
# default base
<foo> :name "Foo" .
# change it
@base <http://example.com/doc/> .
<bar> :name "Bar" .
"""
        g = Graph()
        g.parse(data=input, publicID="http://blah.com/", format="n3")
        print(list(g))
        assert (URIRef("http://blah.com/foo"), None, Literal("Foo")) in g
        assert (URIRef("http://example.com/doc/bar"), None, None) in g

    def test_base_serialize(self):
        g = Graph()
        g.add(
            (
                URIRef("http://example.com/people/Bob"),
                URIRef("urn:knows"),
                URIRef("http://example.com/people/Linda"),
            )
        )
        s = g.serialize(base="http://example.com/", format="n3", encoding="latin-1")
        assert b"<people/Bob>" in s
        g2 = ConjunctiveGraph()
        g2.parse(data=s, publicID="http://example.com/", format="n3")
        assert list(g) == list(g2)

    def test_issue23(self):
        input = """<http://example.com/article1> <http://example.com/title> "this word is in \\u201Cquotes\\u201D"."""

        g = Graph()
        g.parse(data=input, format="n3")

        # Note difference in case of hex code, cwm allows lower-case
        input = """<http://example.com/article1> <http://example.com/title> "this word is in \\u201cquotes\\u201d"."""

        g.parse(data=input, format="n3")

    def test_issue29(self):
        input = """@prefix foo-bar: <http://example.org/> .

foo-bar:Ex foo-bar:name "Test" . """

        g = Graph()
        g.parse(data=input, format="n3")

    def test_issue68(self):
        input = """@prefix : <http://some.url/pome#>.\n\n:Brecon a :Place;\n\t:hasLord\n\t\t:Bernard_of_Neufmarch\xc3\xa9 .\n """

        g = Graph()
        g.parse(data=input, format="n3")

    def test_issue156(self):
        """
        Make sure n3 parser does not choke on UTF-8 BOM
        """
        g = Graph()
        n3_path = os.path.relpath(os.path.join(TEST_DIR, "data/issue156.n3", os.curdir))
        g.parse(n3_path, format="n3")

    def test_issue999(self):
        """
        Make sure the n3 parser does recognize exponent and leading dot in ".171e-11"
        """
        data = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

<http://qudt.org/vocab/unit/MilliM-PER-YR>
  a <http://qudt.org/schema/qudt/Unit> ;
  <http://qudt.org/schema/qudt/conversionMultiplier> .171e-11 ;
  <http://qudt.org/schema/qudt/conversionOffset> 0e+00 ;
  <http://qudt.org/schema/qudt/description> "0.001-fold of the SI base unit metre divided by the unit year" ;
  <http://qudt.org/schema/qudt/hasQuantityKind> <http://qudt.org/vocab/quantitykind/Velocity> ;
  <http://qudt.org/schema/qudt/iec61360Code> "0112/2///62720#UAA868" ;
  <http://qudt.org/schema/qudt/uneceCommonCode> "H66" ;
  rdfs:isDefinedBy <http://qudt.org/2.1/vocab/unit> ;
  rdfs:isDefinedBy <http://qudt.org/vocab/unit> ;
  rdfs:label "MilliM PER YR" ;
  <http://www.w3.org/2004/02/skos/core#prefLabel> "millimetre per year" ;
.
        """
        g = Graph()
        g.parse(data=data, format="n3")
        g.parse(data=data, format="turtle")

    def test_dot_in_prefix(self):
        g = Graph()
        g.parse(
            data="@prefix a.1: <http://example.org/> .\n a.1:cake <urn:x> <urn:y> . \n",
            format="n3",
        )

    def test_model(self):
        g = ConjunctiveGraph()
        g.parse(data=test_data, format="n3")
        i = 0
        for s, p, o in g:
            if isinstance(s, Graph):
                i += 1
        assert i == 3
        assert len(list(g.contexts())) == 13

        g.close()

    def test_quoted_serialization(self):
        g = ConjunctiveGraph()
        g.parse(data=test_data, format="n3")
        g.serialize(format="n3")

    def test_parse(self):
        g = ConjunctiveGraph()
        try:
            g.parse(
                "http://groups.csail.mit.edu/dig/2005/09/rein/examples/troop42-policy.n3",
                format="n3",
            )
        except URLError:
            pytest.skip("No network to retrieve the information, skipping test")

    def test_single_quoted_literals(self):
        test_data = [
            """@prefix : <#> . :s :p 'o' .""",
            """@prefix : <#> . :s :p '''o''' .""",
        ]

        for data in test_data:
            # N3 doesn't accept single quotes around string literals
            g = ConjunctiveGraph()
            with pytest.raises(BadSyntax):
                g.parse(data=data, format="n3")

            g = ConjunctiveGraph()
            g.parse(data=data, format="turtle")
            assert len(g) == 1
            for _, _, o in g:
                assert o == Literal("o")

    def test_empty_prefix(self):
        # this is issue https://github.com/RDFLib/rdflib/issues/312
        g1 = Graph()
        g1.parse(data=":a :b :c .", format="n3")

        g2 = Graph()
        g2.parse(data="@prefix : <#> . :a :b :c .", format="n3")

        assert set(g1) == set(
            g2
        ), "Document with declared empty prefix must match default #"


class TestRegularExpressions:
    def test_exponents(self):
        signs = ("", "+", "-")
        mantissas = (
            "1",
            "1.",
            ".1",
            "12",
            "12.",
            "1.2",
            ".12",
            "123",
            "123.",
            "12.3",
            "1.23",
            ".123",
        )
        es = "eE"
        exps = ("1", "12", "+1", "-1", "+12", "-12")
        for parts in itertools.product(signs, mantissas, es, exps):
            expstring = "".join(parts)
            assert exponent_syntax.match(expstring)

    def test_invalid_exponents(self):
        # Add test cases as needed
        invalid = (".e1",)
        for expstring in invalid:
            assert not exponent_syntax.match(expstring)



from rdflib import RDFS, Graph, Literal, URIRef

_graph_with_label = Graph()
_graph_with_label.add(
    (URIRef("http://example.com/something"), RDFS.label, Literal("Some label"))
)


def test_select_star_sub_select():
    """
    This tests the fix for a bug which returned no results when using `SELECT *` in the
      parent of a sub-select using `SELECT *`.
    """
    results = list(
        _graph_with_label.query(
            """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

            SELECT *
            WHERE {
                {
                    SELECT *
                    WHERE {
                        [] rdfs:label ?label.
                    }
                }
            }
            """
        )
    )

    assert len(results) == 1
    assert results[0].asdict() == {"label": Literal("Some label")}


def test_select_star_multiple_sub_select_star():
    """
    Ensure that we can define select * in multiple sub-selects and still select * (all)
      of the variables out in the parent.
    """
    results = list(
        _graph_with_label.query(
            """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

            SELECT *
            WHERE {
                {
                    SELECT *
                    WHERE {
                        [] rdfs:label ?label.
                    }
                }

                {
                    SELECT *
                    WHERE {
                        [] rdfs:label ?label2.
                    }
                }
            }
            """
        )
    )

    assert len(results) == 1
    assert results[0].asdict() == {
        "label": Literal("Some label"),
        "label2": Literal("Some label"),
    }


def test_select_star_multiple_sub_select_mixed_projections():
    """
    Ensure that we can define select * from one sub-select and define
    projected variables on another sub-select and still select * out of the parent.
    """
    results = list(
        _graph_with_label.query(
            """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

            SELECT *
            WHERE {
                {
                    SELECT *
                    WHERE {
                        [] rdfs:label ?label.
                    }
                }

                {
                    SELECT ?label2
                    WHERE {
                        [] rdfs:label ?label2.
                    }
                }
            }
            """
        )
    )

    assert len(results) == 1
    assert results[0].asdict() == {
        "label": Literal("Some label"),
        "label2": Literal("Some label"),
    }


def test_select_star_multiple_sub_select_defined_projections():
    """
    Ensure that we can define select * from multiple sub-selects which define
    projected variables.
    """
    results = list(
        _graph_with_label.query(
            """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

            SELECT *
            WHERE {
                {
                    SELECT ?label
                    WHERE {
                        [] rdfs:label ?label.
                    }
                }

                {
                    SELECT ?label2
                    WHERE {
                        [] rdfs:label ?label2.
                    }
                }
            }
            """
        )
    )

    assert len(results) == 1
    assert results[0].asdict() == {
        "label": Literal("Some label"),
        "label2": Literal("Some label"),
    }



#!/usr/bin/env python
from __future__ import annotations

import argparse
import enum
import logging
import os
import random
import re
import shutil
import string
import sys
import tarfile
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from tarfile import TarFile, TarInfo
from tempfile import TemporaryDirectory, mkdtemp
from typing import IO, Generator, List, Pattern, Union
from urllib.request import Request, urlopen
from zipfile import ZipFile, ZipInfo

DATA_PATH = Path(__file__).parent


@dataclass
class Resource:
    remote: Union[str, Request]
    local_path: Path

    def fetch(self, tmp_path: Path) -> None:
        raise NotImplementedError()


@dataclass
class FileResource(Resource):
    def fetch(self, tmp_path: Path) -> None:
        if self.local_path.exists():
            logging.debug("info %s", self.local_path)
            os.remove(self.local_path)

        with ExitStack() as xstack:
            request = (
                self.remote
                if isinstance(self.remote, Request)
                else Request(self.remote)
            )
            response = urlopen(request)
            remote_io: IO[bytes] = xstack.enter_context(response)

            local_io = xstack.enter_context(self.local_path.open("wb+"))
            shutil.copyfileobj(remote_io, local_io)

        logging.info("Downloaded %s to %s", request.full_url, self.local_path)


class ArchiveType(enum.Enum):
    ZIP = "zip"
    TAR_GZ = "tar.gz"


@dataclass
class ArchiveResource(Resource):
    type: ArchiveType
    pattern: Pattern[str]
    clean_local: bool = True

    def fetch(self, tmp_path: Path) -> None:
        if self.clean_local and self.local_path.exists():
            logging.debug("info %s", self.local_path)
            shutil.rmtree(self.local_path)
        with ExitStack() as xstack:
            request = (
                self.remote
                if isinstance(self.remote, Request)
                else Request(self.remote)
            )
            response = urlopen(request)
            remote_io: IO[bytes] = xstack.enter_context(response)
            name = (
                "".join(
                    random.choices(
                        string.ascii_uppercase + string.digits + string.ascii_lowercase,
                        k=10,
                    )
                )
                + f".{self.type.value}"
            )
            tmp_file = tmp_path / name
            logging.info("fetching %s to temp file %s", self.remote, tmp_file)
            with tmp_file.open("wb+") as tmp_io:
                shutil.copyfileobj(remote_io, tmp_io)

            archive_file: Union[ZipFile, TarFile]
            if self.type is ArchiveType.ZIP:
                archive_file = xstack.enter_context(ZipFile(tmp_file))
            elif self.type is ArchiveType.TAR_GZ:
                archive_file = xstack.enter_context(tarfile.open(tmp_file, mode="r:gz"))
                # archive_file = xstack.enter_context(TarFile(tmp_file, mode="r|gz"))
            else:
                raise ValueError(f"invalid type {self.type}")

            for member_info in self._member_list(archive_file):
                member_filename = self._member_filename(member_info)
                if self._member_isdir(member_info):
                    logging.debug("Ignoring directory %s", member_filename)
                    continue

                match = self.pattern.match(member_filename)
                if match is None:
                    logging.debug("Ignoring unmatched %s", member_filename)
                    continue
                groups = match.groups()
                if len(groups) > 0:
                    dest_filename = groups[0]

                member_io: IO[bytes]
                with self._member_io(archive_file, member_info) as member_io:
                    local_file = self.local_path / dest_filename
                    if not local_file.parent.exists():
                        local_file.parent.mkdir(parents=True)
                    logging.debug("writing %s to %s", member_filename, local_file)
                    local_file.write_bytes(member_io.read())

        logging.info(
            "Downloaded %s and extracted files matching %s to %s",
            request.full_url,
            self.pattern,
            self.local_path,
        )

        patch_dir = self.local_path.parent / f"{self.local_path.name}.patch"
        if patch_dir.exists():
            logging.info(
                "merging patch content from %s into %s", patch_dir, self.local_path
            )
            for child in patch_dir.glob("**/*"):
                if child.is_dir():
                    logging.debug("ignoring directory %s", child)
                    continue
                if child.name == "README.md" or child.name.endswith(".patch"):
                    logging.debug("ignoring special %s", child)
                    continue
                rel_child = child.relative_to(patch_dir)
                dest = self.local_path / rel_child
                logging.info("copying patch content %s to %s", child, dest)
                shutil.copy2(child, dest)

    @classmethod
    def _member_list(
        cls, archive: Union[ZipFile, TarFile]
    ) -> Union[List[ZipInfo], List[TarInfo]]:
        if isinstance(archive, ZipFile):
            return archive.infolist()
        return archive.getmembers()

    @classmethod
    def _member_isdir(cls, member_info: Union[ZipInfo, TarInfo]) -> bool:
        if isinstance(member_info, ZipInfo):
            return member_info.is_dir()
        return member_info.isdir()

    @classmethod
    def _member_filename(cls, member_info: Union[ZipInfo, TarInfo]) -> str:
        if isinstance(member_info, ZipInfo):
            return member_info.filename
        return member_info.name

    @classmethod
    @contextmanager
    def _member_io(
        cls, archive: Union[ZipFile, TarFile], member_info: Union[ZipInfo, TarInfo]
    ) -> Generator[IO[bytes], None, None]:
        if isinstance(archive, ZipFile):
            assert isinstance(member_info, ZipInfo)
            with archive.open(member_info) as member_io:
                yield member_io
        else:
            assert isinstance(member_info, TarInfo)
            opt_io = archive.extractfile(member_info)
            assert opt_io is not None
            yield opt_io


RESOURCES: List[Resource] = [
    ArchiveResource(
        remote="https://github.com/w3c/N3/archive/c44d123c5958ca04117e28ca3769e2c0820f72e6.zip",
        local_path=(DATA_PATH / "suites" / "w3c" / "n3"),
        type=ArchiveType.ZIP,
        pattern=re.compile(r"^[^\/]+[\/]tests[\/](.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2013/TurtleTests/TESTS.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "turtle"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^[^\/]+[\/](.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2013/N-QuadsTests/TESTS.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "nquads"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^(.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2013/N-TriplesTests/TESTS.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "ntriples"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^(.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2013/TrigTests/TESTS.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "trig"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^(.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2013/RDFXMLTests/TESTS.zip",
        local_path=(DATA_PATH / "suites" / "w3c" / "rdf-xml"),
        type=ArchiveType.ZIP,
        pattern=re.compile(r"^(.+)$"),
    ),
    ArchiveResource(
        # This is taken from a specific git commit instead of the published
        # test suite as the tests are non-normative and not considered part of
        # the test suite, so they could potentially be removed.
        remote="https://github.com/w3c/rdf-tests/archive/d2cc355bf601d8574116f3ee76ca570925f35ac3.zip",
        local_path=(DATA_PATH / "suites" / "w3c" / "rdf-xml-non-normative"),
        type=ArchiveType.ZIP,
        # Not cleaning local directory as it has a constructed manifest.
        clean_local=False,
        pattern=re.compile(
            r"^[^\/]+[\/]rdf-xml[\/]((?:rdfms-empty-property-elements/error003|rdfms-empty-property-elements/test003|rdfms-empty-property-elements/test009|rdfms-xmllang/test001|rdfms-xmllang/test002|rdfms-xml-literal-namespaces/test001|rdfms-xml-literal-namespaces/test002)[.][^.]+)$"
        ),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2009/sparql/docs/tests/sparql11-test-suite-20121023.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "sparql11"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^[^\/]+[\/](.+)$"),
    ),
    ArchiveResource(
        remote="https://www.w3.org/2001/sw/DataAccess/tests/data-r2.tar.gz",
        local_path=(DATA_PATH / "suites" / "w3c" / "dawg-data-r2"),
        type=ArchiveType.TAR_GZ,
        pattern=re.compile(r"^[^\/]+[\/]data-r2[\/](.+)$"),
    ),
    FileResource(
        remote=Request(
            "http://www.w3.org/2000/01/rdf-schema#", headers={"Accept": "text/turtle"}
        ),
        local_path=(DATA_PATH / "defined_namespaces/rdfs.ttl"),
    ),
    FileResource(
        remote=Request(
            "http://www.w3.org/2000/01/rdf-schema#",
            headers={"Accept": "application/rdf+xml"},
        ),
        local_path=(DATA_PATH / "defined_namespaces/rdfs.rdf"),
    ),
    FileResource(
        remote=Request("http://www.w3.org/ns/adms.rdf"),
        local_path=(DATA_PATH / "defined_namespaces/adms.rdf"),
    ),
    FileResource(
        remote=Request("http://www.w3.org/ns/adms.ttl"),
        local_path=(DATA_PATH / "defined_namespaces/adms.ttl"),
    ),
    FileResource(
        remote=Request("https://www.w3.org/ns/rdftest.ttl"),
        local_path=(DATA_PATH / "defined_namespaces/rdftest.ttl"),
    ),
    FileResource(
        remote=Request("https://www.w3.org/2001/sw/DataAccess/tests/test-manifest#"),
        local_path=(DATA_PATH / "defined_namespaces/mf.ttl"),
    ),
    FileResource(
        remote=Request("https://www.w3.org/2001/sw/DataAccess/tests/test-dawg#"),
        local_path=(DATA_PATH / "defined_namespaces/dawgt.ttl"),
    ),
    FileResource(
        remote=Request("https://www.w3.org/2001/sw/DataAccess/tests/test-query#"),
        local_path=(DATA_PATH / "defined_namespaces/qt.ttl"),
    ),
    FileResource(
        remote=Request("https://www.w3.org/2009/sparql/docs/tests/test-update.n3"),
        local_path=(DATA_PATH / "defined_namespaces/ut.n3"),
    ),
    FileResource(
        remote=Request(
            "https://github.com/web-platform-tests/wpt/raw/9d13065419df90d2ad71f3c6b78cc12e7800dae4/html/syntax/parsing/html5lib_tests1.html"
        ),
        local_path=(DATA_PATH / "html5lib_tests1.html"),
    ),
]


@dataclass
class Application:
    parser: argparse.ArgumentParser = field(
        default_factory=lambda: argparse.ArgumentParser(add_help=True)
    )

    def __post_init__(self) -> None:
        parser = self.parser
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            dest="verbosity",
            help="increase verbosity level",
        )
        parser.add_argument(
            "--keep-tmp",
            action="store_true",
            default=False,
        )
        parser.add_argument("paths", nargs="*", type=str)
        parser.set_defaults(handler=self.handle)

    def run(self, args: List[str]) -> None:
        parse_result = self.parser.parse_args(args)

        verbosity = parse_result.verbosity
        if verbosity is not None:
            root_logger = logging.getLogger("")
            root_logger.propagate = True
            new_level = (
                root_logger.getEffectiveLevel()
                - (min(1, verbosity)) * 10
                - min(max(0, verbosity - 1), 9) * 1
            )
            root_logger.setLevel(new_level)

        logging.debug(
            "args = %s, parse_result = %s, logging.level = %s",
            args,
            parse_result,
            logging.getLogger("").getEffectiveLevel(),
        )

        parse_result.handler(parse_result)

    def handle(self, parse_result: argparse.Namespace) -> None:
        logging.debug("entry ...")

        paths = {Path(path).absolute() for path in parse_result.paths}

        logging.debug("paths = %s", paths)

        if parse_result.keep_tmp:
            tmp_path = Path(mkdtemp())
        else:
            tmp_dir = TemporaryDirectory()
            tmp_path = Path(tmp_dir.name)

        for resource in RESOURCES:
            if paths:
                include = False
                for path in paths:
                    try:
                        resource.local_path.absolute().relative_to(path)
                        include = True
                    except ValueError:
                        # not relative to, ignoring
                        pass
                if not include:
                    logging.info("skipping %s", resource.local_path)
                    continue
            resource.fetch(tmp_path)


def main() -> None:
    logging.basicConfig(
        level=os.environ.get("PYLOGGING_LEVEL", logging.INFO),
        stream=sys.stderr,
        datefmt="%Y-%m-%dT%H:%M:%S",
        format=(
            "%(asctime)s.%(msecs)03d %(process)d %(thread)d %(levelno)03d:%(levelname)-8s "
            "%(name)-12s %(module)s:%(lineno)s:%(funcName)s %(message)s"
        ),
    )

    Application().run(sys.argv[1:])


if __name__ == "__main__":
    main()



# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
This test-set demonstrates usage of identifier prefixing and the
forward-slash character in Turtle, JSON-LD, and SPARQL.  The motivating
use case originated with attempts to interact with IANA Media Types as
prefixed concepts, e.g. "application/json" somehow being
"mime:application/json".
"""
from __future__ import annotations

from typing import Set

import pytest

from rdflib import Graph
from rdflib.plugins.sparql.processor import prepareQuery
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import ResultRow
from test.data import TEST_DATA_DIR
from test.utils.graph import cached_graph

query_string_expanded = r"""
SELECT ?nIndividual
WHERE {
  ?nIndividual a <http://example.org/ontology/core/MyClassB> .
}"""

# NOTE: This is expected to fail.  The SPARQL grammar production rules
# for prefixed IRIs, especially at production rule PN_LOCAL, have no way
# to reach the forward-slash or backslash characters.
# https://www.w3.org/TR/rdf-sparql-query/#grammar
query_string_prefixed = r"""
PREFIX ex: <http://example.org/ontology/>
SELECT ?nIndividual
WHERE {
  # NOTE: Syntax is incorrect - forward slash cannot be included in
  # local component of name.
  ?nIndividual a ex:core\/MyClassB .
}"""

PN_LOCAL_BACKSLASH_XFAIL_REASON = """
    Contrary to the ratified SPARQL 1.1 grammar, the RDFlib SPARQL propcessor
    accepts backslashes as part of PN_LOCAL which it treats as escape
    characters.

    There should be a way to instruct the SPARQL parser to operate in strict
    mode, and in strict mode backslashes should not be permitted in PN_LOCAL.

    See https://github.com/RDFLib/rdflib/issues/1871
"""


def _test_query_prepares(query_string: str) -> None:
    """
    Confirm parse behavior of SPARQL engine when a concept would be
    prefixed at a point that leaves a forward-slash character in the
    suffix.
    """
    nsdict = {
        "ex": "http://example.org/ontology/",
        "kb": "http://example.org/kb/",
        "owl": "http://www.w3.org/2002/07/owl#",
    }
    # TODO: A 'strict' flag for prepareQuery is under consideration to
    # adjust parse behavior around backslash characters.
    query_object = prepareQuery(query_string, initNs=nsdict)
    assert isinstance(query_object, Query)


def test_query_prepares_expanded() -> None:
    _test_query_prepares(query_string_expanded)


@pytest.mark.xfail(reason=PN_LOCAL_BACKSLASH_XFAIL_REASON)
def test_query_prepares_prefixed() -> None:
    with pytest.raises(ValueError):
        _test_query_prepares(query_string_prefixed)


def _test_escapes_and_query(
    graph: Graph, query_string: str, expected_query_compiled: bool
) -> None:
    """
    Confirm search-results behavior of SPARQL engine when a concept
    would be prefixed at a point that leaves a forward-slash character
    in the suffix.

    Note that _test_query_prepares also exercises the expected parse
    failure.  This parameterized test is more for demonstrating that
    searching can work without prefixes.
    """
    expected: Set[str] = {
        "http://example.org/kb/individual-b",
    }
    computed: Set[str] = set()

    query_compiled: bool = False
    try:
        query_object = prepareQuery(query_string)
        query_compiled = True
    except Exception:
        pass
    assert expected_query_compiled == query_compiled

    for result in graph.query(query_object):
        assert isinstance(result, ResultRow)
        computed.add(str(result[0]))

    assert expected == computed


def test_escapes_and_query_turtle_expanded() -> None:
    graph = cached_graph((TEST_DATA_DIR / "variants/forward_slash.ttl",))
    _test_escapes_and_query(graph, query_string_expanded, True)


@pytest.mark.xfail(reason=PN_LOCAL_BACKSLASH_XFAIL_REASON, raises=AssertionError)
def test_escapes_and_query_turtle_prefixed() -> None:
    graph = cached_graph((TEST_DATA_DIR / "variants/forward_slash.ttl",))
    _test_escapes_and_query(graph, query_string_prefixed, False)


def test_escapes_and_query_jsonld_expanded() -> None:
    graph = cached_graph((TEST_DATA_DIR / "variants/forward_slash.jsonld",))
    _test_escapes_and_query(graph, query_string_expanded, True)


@pytest.mark.xfail(reason=PN_LOCAL_BACKSLASH_XFAIL_REASON, raises=AssertionError)
def test_escapes_and_query_jsonld_prefixed() -> None:
    graph = cached_graph((TEST_DATA_DIR / "variants/forward_slash.jsonld",))
    _test_escapes_and_query(graph, query_string_prefixed, False)



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class VOID(DefinedNamespace):
    """
    Vocabulary of Interlinked Datasets (VoID)

    The Vocabulary of Interlinked Datasets (VoID) is an RDF Schema vocabulary for expressing metadata about RDF
    datasets. It is intended as a bridge between the publishers and users of RDF data, with applications ranging
    from data discovery to cataloging and archiving of datasets. This document provides a formal definition of the
    new RDF classes and properties introduced for VoID. It is a companion to the main specification document for
    VoID, <em><a href="http://www.w3.org/TR/void/">Describing Linked Datasets with the VoID Vocabulary</a></em>.

    Generated from: http://rdfs.org/ns/void#
    Date: 2020-05-26 14:20:11.911298

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    classPartition: URIRef  # A subset of a void:Dataset that contains only the entities of a certain rdfs:Class.
    classes: URIRef  # The total number of distinct classes in a void:Dataset. In other words, the number of distinct resources occurring as objects of rdf:type triples in the dataset.
    dataDump: URIRef  # An RDF dump, partial or complete, of a void:Dataset.
    distinctObjects: URIRef  # The total number of distinct objects in a void:Dataset. In other words, the number of distinct resources that occur in the object position of triples in the dataset. Literals are included in this count.
    distinctSubjects: URIRef  # The total number of distinct subjects in a void:Dataset. In other words, the number of distinct resources that occur in the subject position of triples in the dataset.
    documents: URIRef  # The total number of documents, for datasets that are published as a set of individual documents, such as RDF/XML documents or RDFa-annotated web pages. Non-RDF documents, such as web pages in HTML or images, are usually not included in this count. This property is intended for datasets where the total number of triples or entities is hard to determine. void:triples or void:entities should be preferred where practical.
    entities: (
        URIRef  # The total number of entities that are described in a void:Dataset.
    )
    exampleResource: URIRef  # example resource of dataset
    feature: URIRef  # feature
    inDataset: URIRef  # Points to the void:Dataset that a document is a part of.
    linkPredicate: URIRef  # a link predicate
    objectsTarget: URIRef  # The dataset describing the objects of the triples contained in the Linkset.
    openSearchDescription: URIRef  # An OpenSearch description document for a free-text search service over a void:Dataset.
    properties: URIRef  # The total number of distinct properties in a void:Dataset. In other words, the number of distinct resources that occur in the predicate position of triples in the dataset.
    property: URIRef  # The rdf:Property that is the predicate of all triples in a property-based partition.
    propertyPartition: URIRef  # A subset of a void:Dataset that contains only the triples of a certain rdf:Property.
    rootResource: URIRef  # A top concept or entry point for a void:Dataset that is structured in a tree-like fashion. All resources in a dataset can be reached by following links from its root resources in a small number of steps.
    sparqlEndpoint: URIRef  # has a SPARQL endpoint at
    subjectsTarget: URIRef  # The dataset describing the subjects of triples contained in the Linkset.
    subset: URIRef  # has subset
    target: URIRef  # One of the two datasets linked by the Linkset.
    triples: URIRef  # The total number of triples contained in a void:Dataset.
    uriLookupEndpoint: (
        URIRef  # Defines a simple URI look-up protocol for accessing a dataset.
    )
    uriRegexPattern: (
        URIRef  # Defines a regular expression pattern matching URIs in the dataset.
    )
    uriSpace: URIRef  # A URI that is a common string prefix of all the entity URIs in a void:Dataset.
    vocabulary: URIRef  # A vocabulary that is used in the dataset.

    # http://www.w3.org/2000/01/rdf-schema#Class
    Dataset: URIRef  # A set of RDF triples that are published, maintained or aggregated by a single provider.
    DatasetDescription: URIRef  # A web resource whose foaf:primaryTopic or foaf:topics include void:Datasets.
    Linkset: URIRef  # A collection of RDF links between two void:Datasets.
    TechnicalFeature: URIRef  # A technical feature of a void:Dataset, such as a supported RDF serialization format.

    # Valid non-python identifiers
    _extras = ["class"]

    _NS = Namespace("http://rdfs.org/ns/void#")



"""
A collection of utilities for canonicalizing and inspecting graphs.

Among other things, they solve of the problem of deterministic bnode
comparisons.

Warning: the time to canonicalize bnodes may increase exponentially on
degenerate larger graphs. Use with care!

Example of comparing two graphs::

    >>> g1 = Graph().parse(format='n3', data='''
    ...     @prefix : <http://example.org/ns#> .
    ...     <http://example.org> :rel
    ...         <http://example.org/same>,
    ...         [ :label "Same" ],
    ...         <http://example.org/a>,
    ...         [ :label "A" ] .
    ... ''')
    >>> g2 = Graph().parse(format='n3', data='''
    ...     @prefix : <http://example.org/ns#> .
    ...     <http://example.org> :rel
    ...         <http://example.org/same>,
    ...         [ :label "Same" ],
    ...         <http://example.org/b>,
    ...         [ :label "B" ] .
    ... ''')
    >>>
    >>> iso1 = to_isomorphic(g1)
    >>> iso2 = to_isomorphic(g2)

These are not isomorphic::

    >>> iso1 == iso2
    False

Diff the two graphs::

    >>> in_both, in_first, in_second = graph_diff(iso1, iso2)

Present in both::

    >>> def dump_nt_sorted(g):
    ...     for l in sorted(g.serialize(format='nt').splitlines()):
    ...         if l: print(l.decode('ascii'))

    >>> dump_nt_sorted(in_both) #doctest: +SKIP
    <http://example.org>
        <http://example.org/ns#rel> <http://example.org/same> .
    <http://example.org>
        <http://example.org/ns#rel> _:cbcaabaaba17fecbc304a64f8edee4335e .
    _:cbcaabaaba17fecbc304a64f8edee4335e
        <http://example.org/ns#label> "Same" .

Only in first::

    >>> dump_nt_sorted(in_first) #doctest: +SKIP
    <http://example.org>
        <http://example.org/ns#rel> <http://example.org/a> .
    <http://example.org>
        <http://example.org/ns#rel> _:cb124e4c6da0579f810c0ffe4eff485bd9 .
    _:cb124e4c6da0579f810c0ffe4eff485bd9
        <http://example.org/ns#label> "A" .

Only in second::

    >>> dump_nt_sorted(in_second) #doctest: +SKIP
    <http://example.org>
        <http://example.org/ns#rel> <http://example.org/b> .
    <http://example.org>
        <http://example.org/ns#rel> _:cb558f30e21ddfc05ca53108348338ade8 .
    _:cb558f30e21ddfc05ca53108348338ade8
        <http://example.org/ns#label> "B" .
"""

from __future__ import annotations

# TODO:
# - Doesn't handle quads.
# - Add warning and/or safety mechanism before working on large graphs?
# - use this in existing Graph.isomorphic?

__all__ = [
    "IsomorphicGraph",
    "to_isomorphic",
    "isomorphic",
    "to_canonical_graph",
    "graph_diff",
    "similar",
]

from collections import defaultdict
from datetime import datetime
from hashlib import sha256
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from rdflib.graph import ConjunctiveGraph, Graph, ReadOnlyGraphAggregate, _TripleType
from rdflib.term import BNode, IdentifiedNode, Node, URIRef

if TYPE_CHECKING:
    from _hashlib import HASH


def _total_seconds(td):
    result = td.days * 24 * 60 * 60
    result += td.seconds
    result += td.microseconds / 1000000.0
    return result


class _runtime:  # noqa: N801
    def __init__(self, label):
        self.label = label

    def __call__(self, f):
        if self.label is None:
            self.label = f.__name__ + "_runtime"

        def wrapped_f(*args, **kwargs):
            start = datetime.now()
            result = f(*args, **kwargs)
            if "stats" in kwargs and kwargs["stats"] is not None:
                stats = kwargs["stats"]
                stats[self.label] = _total_seconds(datetime.now() - start)
            return result

        return wrapped_f


class _call_count:  # noqa: N801
    def __init__(self, label):
        self.label = label

    def __call__(self, f):
        if self.label is None:
            self.label = f.__name__ + "_runtime"

        def wrapped_f(*args, **kwargs):
            if "stats" in kwargs and kwargs["stats"] is not None:
                stats = kwargs["stats"]
                if self.label not in stats:
                    stats[self.label] = 0
                stats[self.label] += 1
            return f(*args, **kwargs)

        return wrapped_f


class IsomorphicGraph(ConjunctiveGraph):
    """An implementation of the RGDA1 graph digest algorithm.

    An implementation of RGDA1 (publication below),
    a combination of Sayers & Karp's graph digest algorithm using
    sum and SHA-256 <http://www.hpl.hp.com/techreports/2003/HPL-2003-235R1.pdf>
    and traces <http://pallini.di.uniroma1.it>, an average case
    polynomial time algorithm for graph canonicalization.

    McCusker, J. P. (2015). WebSig: A Digital Signature Framework for the Web.
    Rensselaer Polytechnic Institute, Troy, NY.
    http://gradworks.umi.com/3727015.pdf
    """

    def __init__(self, **kwargs):
        super(IsomorphicGraph, self).__init__(**kwargs)

    def __eq__(self, other):
        """Graph isomorphism testing."""
        if not isinstance(other, IsomorphicGraph):
            return False
        elif len(self) != len(other):
            return False
        return self.internal_hash() == other.internal_hash()

    def __ne__(self, other):
        """Negative graph isomorphism testing."""
        return not self.__eq__(other)

    def __hash__(self):
        return super(IsomorphicGraph, self).__hash__()

    def graph_digest(self, stats=None):
        """Synonym for IsomorphicGraph.internal_hash."""
        return self.internal_hash(stats=stats)

    def internal_hash(self, stats=None):
        """
        This is defined instead of __hash__ to avoid a circular recursion
        scenario with the Memory store for rdflib which requires a hash lookup
        in order to return a generator of triples.
        """
        return _TripleCanonicalizer(self).to_hash(stats=stats)


HashFunc = Callable[[str], int]
ColorItem = Tuple[Union[int, str], URIRef, Union[int, str]]
ColorItemTuple = Tuple[ColorItem, ...]
HashCache = Optional[Dict[ColorItemTuple, str]]
Stats = Dict[str, Union[int, str]]


class Color:
    def __init__(
        self,
        nodes: List[IdentifiedNode],
        hashfunc: HashFunc,
        color: ColorItemTuple = (),
        hash_cache: HashCache = None,
    ):
        if hash_cache is None:
            hash_cache = {}
        self._hash_cache = hash_cache
        self.color = color
        self.nodes = nodes
        self.hashfunc = hashfunc
        self._hash_color = None

    def __str__(self):
        nodes, color = self.key()
        return "Color %s (%s nodes)" % (color, nodes)

    def key(self):
        return (len(self.nodes), self.hash_color())

    def hash_color(self, color: Optional[Tuple[ColorItem, ...]] = None) -> str:
        if color is None:
            color = self.color
        if color in self._hash_cache:
            return self._hash_cache[color]

        def stringify(x):
            if isinstance(x, Node):
                return x.n3()
            else:
                return str(x)

        if isinstance(color, Node):
            return stringify(color)
        value = 0
        for triple in color:
            value += self.hashfunc(" ".join([stringify(x) for x in triple]))
        val: str = "%x" % value
        self._hash_cache[color] = val
        return val

    def distinguish(self, W: Color, graph: Graph):  # noqa: N803
        colors: Dict[str, Color] = {}
        for n in self.nodes:
            new_color: Tuple[ColorItem, ...] = list(self.color)  # type: ignore[assignment]
            for node in W.nodes:
                new_color += [  # type: ignore[operator]
                    (1, p, W.hash_color()) for s, p, o in graph.triples((n, None, node))
                ]
                new_color += [  # type: ignore[operator]
                    (W.hash_color(), p, 3) for s, p, o in graph.triples((node, None, n))
                ]
            new_color = tuple(new_color)
            new_hash_color = self.hash_color(new_color)

            if new_hash_color not in colors:
                c = Color([], self.hashfunc, new_color, hash_cache=self._hash_cache)
                colors[new_hash_color] = c
            colors[new_hash_color].nodes.append(n)
        return colors.values()

    def discrete(self):
        return len(self.nodes) == 1

    def copy(self):
        return Color(
            self.nodes[:], self.hashfunc, self.color, hash_cache=self._hash_cache
        )


_HashT = Callable[[], "HASH"]


class _TripleCanonicalizer:
    def __init__(self, graph: Graph, hashfunc: _HashT = sha256):
        self.graph = graph

        def _hashfunc(s: str):
            h = hashfunc()
            h.update(str(s).encode("utf8"))
            return int(h.hexdigest(), 16)

        self._hash_cache: HashCache = {}
        self.hashfunc = _hashfunc

    def _discrete(self, coloring: List[Color]) -> bool:
        return len([c for c in coloring if not c.discrete()]) == 0

    def _initial_color(self) -> List[Color]:
        """Finds an initial color for the graph.

        Finds an initial color of the graph by finding all blank nodes and
        non-blank nodes that are adjacent. Nodes that are not adjacent to blank
        nodes are not included, as they are a) already colored (by URI or literal)
        and b) do not factor into the color of any blank node.
        """
        bnodes: Set[BNode] = set()
        others = set()
        self._neighbors = defaultdict(set)
        for s, p, o in self.graph:
            nodes = set([s, p, o])
            b = set([x for x in nodes if isinstance(x, BNode)])
            if len(b) > 0:
                others |= nodes - b
                bnodes |= b
                if isinstance(s, BNode):
                    self._neighbors[s].add(o)
                if isinstance(o, BNode):
                    self._neighbors[o].add(s)
                if isinstance(p, BNode):
                    self._neighbors[p].add(s)
                    self._neighbors[p].add(p)
        if len(bnodes) > 0:
            return [Color(list(bnodes), self.hashfunc, hash_cache=self._hash_cache)] + [
                # type error: List item 0 has incompatible type "Union[IdentifiedNode, Literal]"; expected "IdentifiedNode"
                # type error: Argument 3 to "Color" has incompatible type "Union[IdentifiedNode, Literal]"; expected "Tuple[Tuple[Union[int, str], URIRef, Union[int, str]], ...]"
                Color([x], self.hashfunc, x, hash_cache=self._hash_cache)  # type: ignore[list-item, arg-type]
                for x in others
            ]
        else:
            return []

    def _individuate(self, color, individual):
        new_color = list(color.color)
        new_color.append((len(color.nodes),))

        color.nodes.remove(individual)
        c = Color(
            [individual], self.hashfunc, tuple(new_color), hash_cache=self._hash_cache
        )
        return c

    def _get_candidates(self, coloring: List[Color]) -> Iterator[Tuple[Node, Color]]:
        for c in [c for c in coloring if not c.discrete()]:
            for node in c.nodes:
                yield node, c

    def _refine(self, coloring: List[Color], sequence: List[Color]) -> List[Color]:
        sequence = sorted(sequence, key=lambda x: x.key(), reverse=True)
        coloring = coloring[:]
        while len(sequence) > 0 and not self._discrete(coloring):
            W = sequence.pop()  # noqa: N806
            for c in coloring[:]:
                if len(c.nodes) > 1 or isinstance(c.nodes[0], BNode):
                    colors = sorted(
                        c.distinguish(W, self.graph),
                        key=lambda x: x.key(),
                        reverse=True,
                    )
                    coloring.remove(c)
                    coloring.extend(colors)
                    try:
                        si = sequence.index(c)
                        sequence = sequence[:si] + colors + sequence[si + 1 :]
                    except ValueError:
                        sequence = colors[1:] + sequence
        combined_colors: List[Color] = []
        combined_color_map: Dict[str, Color] = dict()
        for color in coloring:
            color_hash = color.hash_color()
            # This is a hash collision, and be combined into a single color for individuation.
            if color_hash in combined_color_map:
                combined_color_map[color_hash].nodes.extend(color.nodes)
            else:
                combined_colors.append(color)
                combined_color_map[color_hash] = color
        return combined_colors

    @_runtime("to_hash_runtime")
    def to_hash(self, stats: Optional[Stats] = None):
        result = 0
        for triple in self.canonical_triples(stats=stats):
            result += self.hashfunc(" ".join([x.n3() for x in triple]))
        if stats is not None:
            stats["graph_digest"] = "%x" % result
        return result

    def _experimental_path(self, coloring: List[Color]) -> List[Color]:
        coloring = [c.copy() for c in coloring]
        while not self._discrete(coloring):
            color = [x for x in coloring if not x.discrete()][0]
            node = color.nodes[0]
            new_color = self._individuate(color, node)
            coloring.append(new_color)
            coloring = self._refine(coloring, [new_color])
        return coloring

    def _create_generator(
        self,
        colorings: List[List[Color]],
        groupings: Optional[Dict[Node, Set[Node]]] = None,
    ) -> Dict[Node, Set[Node]]:
        if not groupings:
            groupings = defaultdict(set)
        for group in zip(*colorings):
            g = set([c.nodes[0] for c in group])
            for n in group:
                g |= groupings[n]
            for n in g:
                groupings[n] = g
        return groupings

    @_call_count("individuations")
    def _traces(
        self,
        coloring: List[Color],
        stats: Optional[Stats] = None,
        depth: List[int] = [0],
    ) -> List[Color]:
        if stats is not None and "prunings" not in stats:
            stats["prunings"] = 0
        depth[0] += 1
        candidates = self._get_candidates(coloring)
        best: List[List[Color]] = []
        best_score = None
        best_experimental_score = None
        last_coloring = None
        generator: Dict[Node, Set[Node]] = defaultdict(set)
        visited: Set[Node] = set()
        for candidate, color in candidates:
            if candidate in generator:
                v = generator[candidate] & visited
                if len(v) > 0:
                    visited.add(candidate)
                    continue
            visited.add(candidate)
            coloring_copy: List[Color] = []
            color_copy = None
            for c in coloring:
                c_copy = c.copy()
                coloring_copy.append(c_copy)
                if c == color:
                    color_copy = c_copy
            new_color = self._individuate(color_copy, candidate)
            coloring_copy.append(new_color)
            refined_coloring = self._refine(coloring_copy, [new_color])
            color_score = tuple([c.key() for c in refined_coloring])
            experimental = self._experimental_path(coloring_copy)
            experimental_score = set([c.key() for c in experimental])
            if last_coloring:
                generator = self._create_generator(  # type: ignore[unreachable]
                    [last_coloring, experimental], generator
                )
            last_coloring = experimental
            if best_score is None or best_score < color_score:  # type: ignore[unreachable]
                best = [refined_coloring]
                best_score = color_score
                best_experimental_score = experimental_score
            elif best_score > color_score:  # type: ignore[unreachable]
                # prune this branch.
                if stats is not None:
                    stats["prunings"] += 1
            elif experimental_score != best_experimental_score:
                best.append(refined_coloring)
            else:
                # prune this branch.
                if stats is not None:
                    stats["prunings"] += 1
        discrete: List[List[Color]] = [x for x in best if self._discrete(x)]
        if len(discrete) == 0:
            best_score = None
            best_depth = None
            for coloring in best:
                d = [depth[0]]
                new_color = self._traces(coloring, stats=stats, depth=d)
                color_score = tuple([c.key() for c in refined_coloring])
                if best_score is None or color_score > best_score:  # type: ignore[unreachable]
                    discrete = [new_color]
                    best_score = color_score
                    best_depth = d[0]
            depth[0] = best_depth  # type: ignore[assignment]
        return discrete[0]

    def canonical_triples(self, stats: Optional[Stats] = None):
        if stats is not None:
            start_coloring = datetime.now()
        coloring = self._initial_color()
        if stats is not None:
            stats["triple_count"] = len(self.graph)
            stats["adjacent_nodes"] = max(0, len(coloring) - 1)
        coloring = self._refine(coloring, coloring[:])
        if stats is not None:
            stats["initial_coloring_runtime"] = _total_seconds(
                datetime.now() - start_coloring
            )
            stats["initial_color_count"] = len(coloring)

        if not self._discrete(coloring):
            depth = [0]
            coloring = self._traces(coloring, stats=stats, depth=depth)
            if stats is not None:
                stats["tree_depth"] = depth[0]
        elif stats is not None:
            stats["individuations"] = 0
            stats["tree_depth"] = 0
        if stats is not None:
            stats["color_count"] = len(coloring)

        bnode_labels: Dict[Node, str] = dict(
            [(c.nodes[0], c.hash_color()) for c in coloring]
        )
        if stats is not None:
            stats["canonicalize_triples_runtime"] = _total_seconds(
                datetime.now() - start_coloring
            )
        for triple in self.graph:
            result = tuple(self._canonicalize_bnodes(triple, bnode_labels))
            yield result

    def _canonicalize_bnodes(
        self,
        triple: _TripleType,
        labels: Dict[Node, str],
    ):
        for term in triple:
            if isinstance(term, BNode):
                yield BNode(value="cb%s" % labels[term])
            else:
                yield term


def to_isomorphic(graph: Graph) -> IsomorphicGraph:
    if isinstance(graph, IsomorphicGraph):
        return graph
    result = IsomorphicGraph()
    if hasattr(graph, "identifier"):
        result = IsomorphicGraph(identifier=graph.identifier)
    result += graph
    return result


def isomorphic(graph1: Graph, graph2: Graph) -> bool:
    """Compare graph for equality.

    Uses an algorithm to compute unique hashes which takes bnodes into account.

    Examples::

        >>> g1 = Graph().parse(format='n3', data='''
        ...     @prefix : <http://example.org/ns#> .
        ...     <http://example.org> :rel <http://example.org/a> .
        ...     <http://example.org> :rel <http://example.org/b> .
        ...     <http://example.org> :rel [ :label "A bnode." ] .
        ... ''')
        >>> g2 = Graph().parse(format='n3', data='''
        ...     @prefix ns: <http://example.org/ns#> .
        ...     <http://example.org> ns:rel [ ns:label "A bnode." ] .
        ...     <http://example.org> ns:rel <http://example.org/b>,
        ...             <http://example.org/a> .
        ... ''')
        >>> isomorphic(g1, g2)
        True

        >>> g3 = Graph().parse(format='n3', data='''
        ...     @prefix : <http://example.org/ns#> .
        ...     <http://example.org> :rel <http://example.org/a> .
        ...     <http://example.org> :rel <http://example.org/b> .
        ...     <http://example.org> :rel <http://example.org/c> .
        ... ''')
        >>> isomorphic(g1, g3)
        False
    """
    gd1 = _TripleCanonicalizer(graph1).to_hash()
    gd2 = _TripleCanonicalizer(graph2).to_hash()
    return gd1 == gd2


def to_canonical_graph(
    g1: Graph, stats: Optional[Stats] = None
) -> ReadOnlyGraphAggregate:
    """Creates a canonical, read-only graph.

    Creates a canonical, read-only graph where all bnode id:s are based on
    deterministical SHA-256 checksums, correlated with the graph contents.
    """
    graph = Graph()
    graph += _TripleCanonicalizer(g1).canonical_triples(stats=stats)
    return ReadOnlyGraphAggregate([graph])


def graph_diff(g1: Graph, g2: Graph) -> Tuple[Graph, Graph, Graph]:
    """Returns three sets of triples: "in both", "in first" and "in second"."""
    # bnodes have deterministic values in canonical graphs:
    cg1 = to_canonical_graph(g1)
    cg2 = to_canonical_graph(g2)
    in_both = cg1 * cg2
    in_first = cg1 - cg2
    in_second = cg2 - cg1
    return (in_both, in_first, in_second)


_MOCK_BNODE = BNode()


def similar(g1: Graph, g2: Graph):
    """Checks if the two graphs are "similar".

    Checks if the two graphs are "similar", by comparing sorted triples where
    all bnodes have been replaced by a singular mock bnode (the
    ``_MOCK_BNODE``).

    This is a much cheaper, but less reliable, alternative to the comparison
    algorithm in ``isomorphic``.
    """
    return all(t1 == t2 for (t1, t2) in _squashed_graphs_triples(g1, g2))


def _squashed_graphs_triples(g1: Graph, g2: Graph):
    for t1, t2 in zip(sorted(_squash_graph(g1)), sorted(_squash_graph(g2))):
        yield t1, t2


def _squash_graph(graph: Graph):
    return (_squash_bnodes(triple) for triple in graph)


def _squash_bnodes(triple):
    return tuple((isinstance(t, BNode) and _MOCK_BNODE) or t for t in triple)



"""
The :class:`~rdflib.resource.Resource` class wraps a
:class:`~rdflib.graph.Graph`
and a resource reference (i.e. a :class:`rdflib.term.URIRef` or
:class:`rdflib.term.BNode`) to support a resource-oriented way of
working with a graph.

It contains methods directly corresponding to those methods of the Graph
interface that relate to reading and writing data. The difference is that a
Resource also binds a resource identifier, making it possible to work without
tracking both the graph and a current subject. This makes for a "resource
oriented" style, as compared to the triple orientation of the Graph API.

Resulting generators are also wrapped so that any resource reference values
(:class:`rdflib.term.URIRef` and :class:`rdflib.term.BNode`) are in turn
wrapped as Resources. (Note that this behaviour differs from the corresponding
methods in :class:`~rdflib.graph.Graph`, where no such conversion takes place.)


Basic Usage Scenario
--------------------

Start by importing things we need and define some namespaces::

    >>> from rdflib import *
    >>> FOAF = Namespace("http://xmlns.com/foaf/0.1/")
    >>> CV = Namespace("http://purl.org/captsolo/resume-rdf/0.2/cv#")

Load some RDF data::

    >>> graph = Graph().parse(format='n3', data='''
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    ... @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
    ... @prefix foaf: <http://xmlns.com/foaf/0.1/> .
    ... @prefix cv: <http://purl.org/captsolo/resume-rdf/0.2/cv#> .
    ...
    ... @base <http://example.org/> .
    ...
    ... </person/some1#self> a foaf:Person;
    ...     rdfs:comment "Just a Python & RDF hacker."@en;
    ...     foaf:depiction </images/person/some1.jpg>;
    ...     foaf:homepage <http://example.net/>;
    ...     foaf:name "Some Body" .
    ...
    ... </images/person/some1.jpg> a foaf:Image;
    ...     rdfs:label "some 1"@en;
    ...     rdfs:comment "Just an image"@en;
    ...     foaf:thumbnail </images/person/some1-thumb.jpg> .
    ...
    ... </images/person/some1-thumb.jpg> a foaf:Image .
    ...
    ... [] a cv:CV;
    ...     cv:aboutPerson </person/some1#self>;
    ...     cv:hasWorkHistory [ cv:employedIn </#company>;
    ...             cv:startDate "2009-09-04"^^xsd:date ] .
    ... ''')

Create a Resource::

    >>> person = Resource(
    ...     graph, URIRef("http://example.org/person/some1#self"))

Retrieve some basic facts::

    >>> person.identifier
    rdflib.term.URIRef('http://example.org/person/some1#self')

    >>> person.value(FOAF.name)
    rdflib.term.Literal('Some Body')

    >>> person.value(RDFS.comment)
    rdflib.term.Literal('Just a Python & RDF hacker.', lang='en')

Resources can be sliced (like graphs, but the subject is fixed)::

    >>> for name in person[FOAF.name]:
    ...     print(name)
    Some Body
    >>> person[FOAF.name : Literal("Some Body")]
    True

Resources as unicode are represented by their identifiers as unicode::

    >>> %(unicode)s(person)  #doctest: +SKIP
    'Resource(http://example.org/person/some1#self'

Resource references are also Resources, so you can easily get e.g. a qname
for the type of a resource, like::

    >>> person.value(RDF.type).qname()
    'foaf:Person'

Or for the predicates of a resource::

    >>> sorted(
    ...     p.qname() for p in person.predicates()
    ... )  #doctest: +NORMALIZE_WHITESPACE +SKIP
    ['foaf:depiction', 'foaf:homepage',
     'foaf:name', 'rdf:type', 'rdfs:comment']

Follow relations and get more data from their Resources as well::

    >>> for pic in person.objects(FOAF.depiction):
    ...     print(pic.identifier)
    ...     print(pic.value(RDF.type).qname())
    ...     print(pic.value(FOAF.thumbnail).identifier)
    http://example.org/images/person/some1.jpg
    foaf:Image
    http://example.org/images/person/some1-thumb.jpg

    >>> for cv in person.subjects(CV.aboutPerson):
    ...     work = list(cv.objects(CV.hasWorkHistory))[0]
    ...     print(work.value(CV.employedIn).identifier)
    ...     print(work.value(CV.startDate))
    http://example.org/#company
    2009-09-04

It's just as easy to work with the predicates of a resource::

    >>> for s, p in person.subject_predicates():
    ...     print(s.value(RDF.type).qname())
    ...     print(p.qname())
    ...     for s, o in p.subject_objects():
    ...         print(s.value(RDF.type).qname())
    ...         print(o.value(RDF.type).qname())
    cv:CV
    cv:aboutPerson
    cv:CV
    foaf:Person

This is useful for e.g. inspection::

    >>> thumb_ref = URIRef("http://example.org/images/person/some1-thumb.jpg")
    >>> thumb = Resource(graph, thumb_ref)
    >>> for p, o in thumb.predicate_objects():
    ...     print(p.qname())
    ...     print(o.qname())
    rdf:type
    foaf:Image


Schema Example
--------------

With this artificial schema data::

    >>> graph = Graph().parse(format='n3', data='''
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    ... @prefix owl: <http://www.w3.org/2002/07/owl#> .
    ... @prefix v: <http://example.org/def/v#> .
    ...
    ... v:Artifact a owl:Class .
    ...
    ... v:Document a owl:Class;
    ...     rdfs:subClassOf v:Artifact .
    ...
    ... v:Paper a owl:Class;
    ...     rdfs:subClassOf v:Document .
    ...
    ... v:Choice owl:oneOf (v:One v:Other) .
    ...
    ... v:Stuff a rdf:Seq; rdf:_1 v:One; rdf:_2 v:Other .
    ...
    ... ''')

From this class::

    >>> artifact = Resource(graph, URIRef("http://example.org/def/v#Artifact"))

we can get at subclasses::

    >>> subclasses = list(artifact.transitive_subjects(RDFS.subClassOf))
    >>> [c.qname() for c in subclasses]
    ['v:Artifact', 'v:Document', 'v:Paper']

and superclasses from the last subclass::

    >>> [c.qname() for c in subclasses[-1].transitive_objects(RDFS.subClassOf)]
    ['v:Paper', 'v:Document', 'v:Artifact']

Get items from the Choice::

    >>> choice = Resource(graph, URIRef("http://example.org/def/v#Choice"))
    >>> [it.qname() for it in choice.value(OWL.oneOf).items()]
    ['v:One', 'v:Other']

On add, other resources are auto-unboxed:
    >>> paper = Resource(graph, URIRef("http://example.org/def/v#Paper"))
    >>> paper.add(RDFS.subClassOf, artifact)
    >>> artifact in paper.objects(RDFS.subClassOf) # checks Resource instance
    True
    >>> (paper._identifier, RDFS.subClassOf, artifact._identifier) in graph
    True


Technical Details
-----------------

Comparison is based on graph and identifier::

    >>> g1 = Graph()
    >>> t1 = Resource(g1, URIRef("http://example.org/thing"))
    >>> t2 = Resource(g1, URIRef("http://example.org/thing"))
    >>> t3 = Resource(g1, URIRef("http://example.org/other"))
    >>> t4 = Resource(Graph(), URIRef("http://example.org/other"))

    >>> t1 is t2
    False

    >>> t1 == t2
    True
    >>> t1 != t2
    False

    >>> t1 == t3
    False
    >>> t1 != t3
    True

    >>> t3 != t4
    True

    >>> t3 < t1 and t1 > t3
    True
    >>> t1 >= t1 and t1 >= t3
    True
    >>> t1 <= t1 and t3 <= t1
    True

    >>> t1 < t1 or t1 < t3 or t3 > t1 or t3 > t3
    False

Hash is computed from graph and identifier::

    >>> g1 = Graph()
    >>> t1 = Resource(g1, URIRef("http://example.org/thing"))

    >>> hash(t1) == hash(Resource(g1, URIRef("http://example.org/thing")))
    True

    >>> hash(t1) == hash(Resource(Graph(), t1.identifier))
    False
    >>> hash(t1) == hash(Resource(Graph(), URIRef("http://example.org/thing")))
    False

The Resource class is suitable as a base class for mapper toolkits. For
example, consider this utility for accessing RDF properties via qname-like
attributes::

    >>> class Item(Resource):
    ...
    ...     def __getattr__(self, p):
    ...         return list(self.objects(self._to_ref(*p.split('_', 1))))
    ...
    ...     def _to_ref(self, pfx, name):
    ...         return URIRef(self._graph.store.namespace(pfx) + name)

It works as follows::

    >>> graph = Graph().parse(format='n3', data='''
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
    ... @prefix foaf: <http://xmlns.com/foaf/0.1/> .
    ...
    ... @base <http://example.org/> .
    ... </person/some1#self>
    ...     foaf:name "Some Body";
    ...     foaf:depiction </images/person/some1.jpg> .
    ... </images/person/some1.jpg> rdfs:comment "Just an image"@en .
    ... ''')

    >>> person = Item(graph, URIRef("http://example.org/person/some1#self"))

    >>> print(person.foaf_name[0])
    Some Body

The mechanism for wrapping references as resources cooperates with subclasses.
Therefore, accessing referenced resources automatically creates new ``Item``
objects::

    >>> isinstance(person.foaf_depiction[0], Item)
    True

    >>> print(person.foaf_depiction[0].rdfs_comment[0])
    Just an image

"""

from rdflib.namespace import RDF
from rdflib.paths import Path
from rdflib.term import BNode, Node, URIRef

__all__ = ["Resource"]


class Resource:
    def __init__(self, graph, subject):
        self._graph = graph
        self._identifier = subject

    @property
    def graph(self):
        return self._graph

    @property
    def identifier(self):
        return self._identifier

    def __hash__(self):
        return hash(Resource) ^ hash(self._graph) ^ hash(self._identifier)

    def __eq__(self, other):
        return (
            isinstance(other, Resource)
            and self._graph == other._graph
            and self._identifier == other._identifier
        )

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Resource):
            return self._identifier < other._identifier
        else:
            return False

    def __gt__(self, other):
        return not (self < other or self == other)

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return not self < other

    def __unicode__(self):
        return str(self._identifier)

    def add(self, p, o):
        if isinstance(o, Resource):
            o = o._identifier

        self._graph.add((self._identifier, p, o))

    def remove(self, p, o=None):
        if isinstance(o, Resource):
            o = o._identifier

        self._graph.remove((self._identifier, p, o))

    def set(self, p, o):
        if isinstance(o, Resource):
            o = o._identifier

        self._graph.set((self._identifier, p, o))

    def subjects(self, predicate=None):  # rev
        return self._resources(self._graph.subjects(predicate, self._identifier))

    def predicates(self, o=None):
        if isinstance(o, Resource):
            o = o._identifier

        return self._resources(self._graph.predicates(self._identifier, o))

    def objects(self, predicate=None):
        return self._resources(self._graph.objects(self._identifier, predicate))

    def subject_predicates(self):
        return self._resource_pairs(self._graph.subject_predicates(self._identifier))

    def subject_objects(self):
        return self._resource_pairs(self._graph.subject_objects(self._identifier))

    def predicate_objects(self):
        return self._resource_pairs(self._graph.predicate_objects(self._identifier))

    def value(self, p=RDF.value, o=None, default=None, any=True):
        if isinstance(o, Resource):
            o = o._identifier

        return self._cast(self._graph.value(self._identifier, p, o, default, any))

    def items(self):
        return self._resources(self._graph.items(self._identifier))

    def transitive_objects(self, predicate, remember=None):
        return self._resources(
            self._graph.transitive_objects(self._identifier, predicate, remember)
        )

    def transitive_subjects(self, predicate, remember=None):
        return self._resources(
            self._graph.transitive_subjects(predicate, self._identifier, remember)
        )

    def qname(self):
        return self._graph.qname(self._identifier)

    def _resource_pairs(self, pairs):
        for s1, s2 in pairs:
            yield self._cast(s1), self._cast(s2)

    def _resource_triples(self, triples):
        for s, p, o in triples:
            yield self._cast(s), self._cast(p), self._cast(o)

    def _resources(self, nodes):
        for node in nodes:
            yield self._cast(node)

    def _cast(self, node):
        if isinstance(node, (BNode, URIRef)):
            return self._new(node)
        else:
            return node

    def __iter__(self):
        return self._resource_triples(
            self._graph.triples((self.identifier, None, None))
        )

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.step:
                raise TypeError(
                    "Resources fix the subject for slicing, and can only be sliced by predicate/object. "
                )
            p, o = item.start, item.stop
            if isinstance(p, Resource):
                p = p._identifier
            if isinstance(o, Resource):
                o = o._identifier
            if p is None and o is None:
                return self.predicate_objects()
            elif p is None:
                return self.predicates(o)
            elif o is None:
                return self.objects(p)
            else:
                return (self.identifier, p, o) in self._graph
        elif isinstance(item, (Node, Path)):
            return self.objects(item)
        else:
            raise TypeError(
                "You can only index a resource by a single rdflib term, a slice of rdflib terms, not %s (%s)"
                % (item, type(item))
            )

    def __setitem__(self, item, value):
        self.set(item, value)

    def _new(self, subject):
        return type(self)(self._graph, subject)

    def __str__(self):
        return "Resource(%s)" % self._identifier

    def __repr__(self):
        return "Resource(%s,%s)" % (self._graph, self._identifier)



from rdflib import Graph

graph = Graph().parse(
    format="n3",
    data="""
@prefix dct: <http://purl.org/dc/terms/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://example.org/doc> a foaf:Document;
    dct:created "2011-03-20"^^xsd:date .
""",
)


class TestPrefixTypes:
    """N3/Turtle serializers should use prefixes,
    also for types and datatypes

    This is issue 161
    http://code.google.com/p/rdflib/issues/detail?id=161
    """

    def test(self):
        s = graph.serialize(format="n3")
        assert "foaf:Document" in s
        assert "xsd:date" in s
